# -*- coding: utf-8 -*-
##############################################################################
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
##############################################################################

import mock
import os
import time

from collections import namedtuple
from slapos.grid.promise import PromiseError
from slapos.promise.plugin.check_network_transit import RunPromise
from . import TestPromisePluginMixin


class TestCheckNetworkTransit(TestPromisePluginMixin):

  promise_name = "monitor-network-transit.py"

  def setUp(self):
    super(TestCheckNetworkTransit, self).setUp()
    self.network_data = namedtuple('network_data', ['bytes_recv', 'bytes_sent'])

  def writePromise(self, **kw):
    super(TestCheckNetworkTransit, self).writePromise(self.promise_name,
      "from %s import %s\nextra_config_dict = %r\n"
      % (RunPromise.__module__, RunPromise.__name__, kw))

  def runPromise(self, summary=None, failed=False):
    self.configureLauncher(enable_anomaly=True, force=True)
    with mock.patch('psutil.net_io_counters', return_value=summary):
      if failed:
        self.assertRaises(PromiseError, self.launcher.run)
      else:
        self.launcher.run()
    result = self.getPromiseResult(self.promise_name)['result']
    self.assertEqual(result['failed'], failed)
    return result['message']

  def test_network_transit_ok(self):
    message = "Network transit OK"
    mock_stats = {'bytes_recv':1e6, 'bytes_sent':1e6}
    self.writePromise(**{
        'transit-period': 1,
    })
    self.runPromise(self.network_data(**{'bytes_recv':1e3, 'bytes_sent':1e3}))
    time.sleep(0.5)
    self.assertEqual(message, self.runPromise(self.network_data(**mock_stats)))

  def test_network_min_nok(self):
    message = "Network congested, data amount over the last 1 seconds"\
      " reached minimum threshold:    1.4K (threshold is  102.4K)"
    mock_stats = {'bytes_recv':1e3, 'bytes_sent':1e3}
    self.writePromise(**{
        'min-data-amount': 0.1, # MB
        'transit-period': 1,
    })
    self.runPromise(self.network_data(**{'bytes_recv':300, 'bytes_sent':300}))
    time.sleep(0.5)
    self.assertEqual(message, self.runPromise(self.network_data(**mock_stats)))

  def test_network_max_nok(self):
    message = "Network congested, data amount over the last 1 seconds"\
      " reached maximum threshold:    1.1M (threshold is    1.0M)"
    mock_stats = {'bytes_recv':0.7e6, 'bytes_sent':0.5e6}
    self.writePromise(**{
        'max-data-amount': 1, # MB
        'transit-period': 1,
    })
    self.runPromise(self.network_data(**{'bytes_recv':300, 'bytes_sent':300}))
    time.sleep(0.5)
    self.assertEqual(message, self.runPromise(self.network_data(**mock_stats)))

  def test_network_transit_nok(self):
    message = "Network congested, data amount over the last 1 seconds reached minimum threshold:    0.0B (threshold is    0.0B)\n"\
      "Network congested, data amount over the last 1 seconds reached maximum threshold:    0.0B (threshold is    0.0B)"
    mock_stats = {'bytes_recv':1e6, 'bytes_sent':1e6}
    self.writePromise(**{
        'last_transit_file':'last_transit_file',
        'max-data-amount': 0,
        'min-data-amount': 0,
        'transit-period': 1,
    })
    self.runPromise(self.network_data(**{'bytes_recv':1e6, 'bytes_sent':1e6}))
    time.sleep(0.5)
    self.assertEqual(message, self.runPromise(self.network_data(**mock_stats)))

if __name__ == '__main__':
  unittest.main()



# coding: utf-8
# Copyright (C) 2025  Nexedi SA and Contributors.
#                     Łukasz Nowak <luke@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import datetime
import gzip
import re
import contextlib
import click

TIMESTAMP_DETECT = re.compile(r'\[(?P<timestamp>.*)\].*')
PROCESSING_DETECT = re.compile(
  r'\[(?P<timestamp>.*)\].* Processing Computer Partition (?P<partition>.*)\.')
ERROR_DETECT = re.compile(
  r'\[(?P<timestamp>.*)\].* ERROR.* Failed to run buildout profile in directo'
  'ry.*')
SOFTWARE_DETECT = re.compile(r'.*Software URL: (?P<software_release>.*)')
VERSION_DETECT = re.compile(r'.*(?P<version>1\.0\.[0-9]+)/.*')


@contextlib.contextmanager
def auto_open(filename, mode):
  if filename.endswith('.gz'):
    with gzip.open(filename, mode) as f:
      yield f
  else:
    with open(filename, mode) as f:
      yield f


@click.command(short_help="Times slapos instance log files")
@click.argument(
  'file_list',
  nargs=-1,
  type=click.Path(),  # can't use click.File, as existence is required and
                      # that increases usage complexity
)
def main(file_list):
  """
  Allows to analyze and time slapos node log files.
  FILE_LIST can be provided in order, then analysis continuity will be kept.
  """
  processing_state = {}
  for filename in file_list:
    with auto_open(filename, mode='rt') as fh:
      for line in fh.readlines():
        line = line.strip()[:500]
        if not line.startswith('['):
          continue
        processing = PROCESSING_DETECT.fullmatch(line)
        if processing:
          run_id = processing.groupdict()['timestamp'].split(',')[0]
          timestamp = datetime.datetime.strptime(
            run_id, '%Y-%m-%d %H:%M:%S').timestamp()
          if 'start' not in processing_state:
            processing_state['run_id'] = run_id
            processing_state['start'] = timestamp
            processing_state['partition'] = processing.groupdict()['partition']
            processing_state['status'] = 'OK'
          else:
            processing_state['end'] = datetime.datetime.strptime(
              processing_state['previous_timestamp'], '%Y-%m-%d %H:%M:%S'
            ).timestamp()
            processing_state['elapsed'] = int(
              processing_state['end'] - processing_state['start'])
            print(
              '%(partition)s;%(run_id)s;%(version)s;%(status)s;%(elapsed)s'
              % processing_state)
            processing_state['run_id'] = run_id
            processing_state['start'] = timestamp
            processing_state['partition'] = processing.groupdict()['partition']
            processing_state['status'] = 'OK'
          continue
        processing_state['previous_timestamp'] = TIMESTAMP_DETECT.fullmatch(
          line).groupdict()['timestamp'].split(',')[0]
        error = ERROR_DETECT.fullmatch(line)
        if error:
          processing_state['status'] = 'ERR'
        software = SOFTWARE_DETECT.fullmatch(line)
        if software:
          software_release = software.groupdict()['software_release']
          version = VERSION_DETECT.fullmatch(software_release)
          if version is not None:
            processing_state['version'] = version['version']
          else:
            processing_state['version'] = software_release
          continue
  else:
    processing_state['end'] = datetime.datetime.strptime(
      processing_state['previous_timestamp'], '%Y-%m-%d %H:%M:%S').timestamp()
    processing_state['elapsed'] = int(
      processing_state['end'] - processing_state['start'])
    if processing_state['status'] != 'ERR':
      processing_state['status'] = 'UNK/LAST'
    print(
      '%(partition)s;%(run_id)s;%(version)s;%(status)s;%(elapsed)s'
      % processing_state)


if __name__ == '__main__':
  main()



# -*- coding: utf-8 -*-
##############################################################################
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
##############################################################################

import os
import time
from datetime import datetime
from datetime import timedelta
from slapos.grid.promise import PromiseError
from slapos.promise.plugin.check_oru_lof import RunPromise
from . import TestPromisePluginMixin


class TestCheckOruLOFSuccess(TestPromisePluginMixin):

  promise_name = "check-oru-lof.py"

  def setUp(self):
    super(TestCheckOruLOFSuccess, self).setUp()
    self.netconf_log = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'netconf.json.log')

  def writePromise(self, **kw):
    super(TestCheckOruLOFSuccess, self).writePromise(self.promise_name,
      "from %s import %s\nextra_config_dict = %r\n"
      % (RunPromise.__module__, RunPromise.__name__, kw))

  def test_promise_success(self):
    with open(self.netconf_log, 'w+') as f:
      f.write("""{"time": "%s", "log_level": "INFO", "message": "", "data": {"notification": {"@xmlns": "urn:ietf:params:xml:ns:netconf:notification:1.0", "eventTime": "1970-01-05T00:38:50Z", "alarm-notif": {"@xmlns": "urn:o-ran:fm:1.0", "fault-id": "1002", "fault-source": "Radio Module", "affected-objects": {"name": "Radio Module"}, "fault-severity": "CRITICAL", "is-cleared": "false", "fault-text": "LOF Alarm", "event-time": "1970-01-05T00:38:50Z"}}}}
{"time": "%s", "log_level": "INFO", "message": "", "data": {"notification": {"@xmlns": "urn:ietf:params:xml:ns:netconf:notification:1.0", "eventTime": "1970-01-05T00:38:50Z", "alarm-notif": {"@xmlns": "urn:o-ran:fm:1.0", "fault-id": "1002", "fault-source": "Radio Module", "affected-objects": {"name": "Radio Module"}, "fault-severity": "CRITICAL", "is-cleared": "true", "fault-text": "LOF Alarm", "event-time": "1970-01-05T00:38:50Z"}}}}""" % (
      (datetime.now() - timedelta(seconds=25)).strftime("%Y-%m-%d %H:%M:%S,%f")[:-3],
      (datetime.now() - timedelta(seconds=15)).strftime("%Y-%m-%d %H:%M:%S,%f")[:-3],
      ))
    self.writePromise(**{
        'netconf-log': self.netconf_log,
    })
    self.configureLauncher()
    self.launcher.run()

  def test_promise_fail(self):
    with open(self.netconf_log, 'w+') as f:
      f.write("""{"time": "%s", "log_level": "INFO", "message": "", "data": {"notification": {"@xmlns": "urn:ietf:params:xml:ns:netconf:notification:1.0", "eventTime": "1970-01-05T00:38:50Z", "alarm-notif": {"@xmlns": "urn:o-ran:fm:1.0", "fault-id": "1002", "fault-source": "Radio Module", "affected-objects": {"name": "Radio Module"}, "fault-severity": "CRITICAL", "is-cleared": "true", "fault-text": "LOF Alarm", "event-time": "1970-01-05T00:38:50Z"}}}}
{"time": "%s", "log_level": "INFO", "message": "", "data": {"notification": {"@xmlns": "urn:ietf:params:xml:ns:netconf:notification:1.0", "eventTime": "1970-01-05T00:38:50Z", "alarm-notif": {"@xmlns": "urn:o-ran:fm:1.0", "fault-id": "1002", "fault-source": "Radio Module", "affected-objects": {"name": "Radio Module"}, "fault-severity": "CRITICAL", "is-cleared": "false", "fault-text": "LOF Alarm", "event-time": "1970-01-05T00:38:50Z"}}}}""" % (
      (datetime.now() - timedelta(seconds=25)).strftime("%Y-%m-%d %H:%M:%S,%f")[:-3],
      (datetime.now() - timedelta(seconds=15)).strftime("%Y-%m-%d %H:%M:%S,%f")[:-3],
      ))
    self.writePromise(**{
        'netconf-log': self.netconf_log,
    })
    self.configureLauncher()
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertEqual("Loss of Frame (LOF) alarm is on, affected objects are: {'name': 'Radio Module'}", self.getPromiseResult(self.promise_name)['result']['message'])



from .util import JSONPromise, get_json_log_data_interval

from zope.interface import implementer
from slapos.grid.promise import interface

@implementer(interface.IPromise)
class RunPromise(JSONPromise):
  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    self.setPeriodicity(float(self.getConfig('frequency', 1)))
    self.amarisoft_rf_info_log = self.getConfig('amarisoft-rf-info-log')
    self.stats_period = int(self.getConfig('stats-period'))

  def sense(self):

      data_list = get_json_log_data_interval(self.amarisoft_rf_info_log, self.stats_period * 2)
      if len(data_list) < 1:
        self.logger.error("rf_info: stale data")
        return

      rf_info_text = data_list[0]['rf_info']

      if 'Sync: gps (locked)'.lower() in rf_info_text.lower():
        self.logger.info("GPS locked")
      else:
        self.logger.error("GPS not locked")


  def test(self):
    """
      Called after sense() if the instance is still converging.
      Returns success or failure based on sense results.

      In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)


  def anomaly(self):
    """
      Called after sense() if the instance has finished converging.
      Returns success or failure based on sense results.
      Failure signals the instance has diverged.

      In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=1, failure_amount=1)



# -*- coding: utf-8 -*-
##############################################################################
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
##############################################################################

import mock
import os
import time
import json
from datetime import datetime
from datetime import timedelta
from slapos.grid.promise import PromiseError
from slapos.promise.plugin.check_gps_lock import RunPromise
from . import TestPromisePluginMixin


class TestCheckGPSLock(TestPromisePluginMixin):

  promise_name = "check-gps-lock.py"

  def setUp(self):
    super(TestCheckGPSLock, self).setUp()
    self.amarisoft_rf_info_log = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'amarisoft_rf_info.json.log')

    rf_info = \
"""
TRX SDR driver 2024-11-20, API v15
PCIe RFIC /dev/sdr0:
  Hardware ID: 0x4b01
  DNA: [0x0070b5443f4a2854]
  Serial: ''
  FPGA revision: 2021-10-08  15:38:12
  FPGA vccint: 1.01 V
  FPGA vccaux: 1.78 V
  FPGA vccbram: 1.01 V
  FPGA temperature: 55.5 °C
  AD9361 temperature: 37 °C
  AGC: Off
  Sync: gps (locked)
  Clock: internal (locked)
  Clock tune: -0.2 ppm
  NUMA: -1
  Caps: 
  TX channels: 2;  RX channels: 2
  DMA: 1 ch, 32 bits, SMem index: Off, RX Headers: Off
  DMA0: TX fifo: 33.33us  Usage=8/12288 (0%)
  DMA0: RX fifo: 33.33us  Usage=8/12288 (0%)
  DMA0: TX_Underflows: 0  RX_Overflows: 0
  BUFS: bps=32 TX idx=9293176/52600.113 (33.3us) RX: idx=9293176/52600.111 (33.3us)

GPS info:
  UTC:  2024-11-29 10:21:54
  pos:  lat=50.64428°  long=3.07739°
  height: 39.1m  nb_sats: 12
"""
    self.rf_info_data = {'message': 'rf', 'rf_info': rf_info}


  def writeLog(self, data, ago=5):
    with open(self.amarisoft_rf_info_log, 'w') as f:
      f.write(
      """{"time": "%s", "log_level": "INFO", "message": "RF info", "data": %s}""" %
        ((datetime.now() - timedelta(seconds=ago)).strftime("%Y-%m-%d %H:%M:%S")[:-3], json.dumps(data)))

  def writePromise(self, **kw):
    kw.update({'amarisoft-rf-info-log': self.amarisoft_rf_info_log,
               'stats-period':          100})
    super(TestCheckGPSLock, self).writePromise(self.promise_name,
      "from %s import %s\nextra_config_dict = %r\n"
      % (RunPromise.__module__, RunPromise.__name__, kw))

  def test_locked_ok(self):
    self.writeLog(self.rf_info_data)
    self.writePromise()
    self.configureLauncher()
    self.launcher.run()

  def test_stale_data(self):
    self.writeLog(self.rf_info_data, ago=500)
    self.writePromise()
    self.configureLauncher()
    with self.assertRaisesRegex(PromiseError, 'rf_info: stale data'):
      self.launcher.run()

if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
##############################################################################
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
##############################################################################

import mock
import os
import time
import json
from datetime import datetime
from datetime import timedelta
from slapos.grid.promise import PromiseError
from slapos.promise.plugin.check_cpri_lock import RunPromise
from . import TestPromisePluginMixin


class TestCheckCpriLock(TestPromisePluginMixin):

  promise_name = "check-cpri-lock.py"

  def setUp(self):
    super(TestCheckCpriLock, self).setUp()
    self.amarisoft_rf_info_log = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'amarisoft_rf_info.json.log')

    rf_info = \
"""
TRX SDR driver 2023-09-07, API v15/18
PCIe CPRI /dev/sdr2@1:
  Hardware ID: 0x4b12
  DNA: [0x0048248a334a7054]
  Serial: ''
  FPGA revision: 2023-06-23  10:05:24
  FPGA vccint: 0.98 V
  FPGA vccaux: 1.76 V
  FPGA vccbram: 0.98 V
  FPGA temperature: 71.9 °C
  Clock tune: 0.0 ppm
  NUMA: 0
  CPRI_option: '5' (x8) lock=no
  DMA0: TX fifo: 66.67us  Usage=16/32768 (0%)
  DMA0: RX fifo: 66.67us  Usage=16/32768 (0%)
  DMA0 Underflows: 0
  DMA0 Overflows: 0
PCIe CPRI /dev/sdr3@1:
  Hardware ID: 0x4b12
  DNA: [0x0048248a334a7054]
  Serial: ''
  FPGA revision: 2023-06-23  10:05:24
  FPGA vccint: 0.98 V
  FPGA vccaux: 1.77 V
  FPGA vccbram: 0.98 V
  FPGA temperature: 71.7 °C
  Clock tune: 0.0 ppm
  NUMA: 0
  CPRI_option: '5' (x8) lock=HW+SW rx/tx=46.606us
    Port #0: T14=46.606us
  DMA0: TX fifo: 66.67us  Usage=16/32768 (0%)
  DMA0: RX fifo: 66.67us  Usage=16/32768 (0%)
  DMA0 Underflows: 0
  DMA0 Overflows: 0
PCIe SDR /dev/sdr4@0:
  AAA: bbb
"""
    self.rf_info_data = {'message': 'rf', 'rf_info': rf_info}


  def writeLog(self, data, ago=5):
    with open(self.amarisoft_rf_info_log, 'w') as f:
      f.write(
      """{"time": "%s", "log_level": "INFO", "message": "RF info", "data": %s}""" %
        ((datetime.now() - timedelta(seconds=ago)).strftime("%Y-%m-%d %H:%M:%S")[:-3], json.dumps(data)))

  def writePromise(self, **kw):
    kw.update({'amarisoft-rf-info-log': self.amarisoft_rf_info_log,
               'stats-period':          100})
    super(TestCheckCpriLock, self).writePromise(self.promise_name,
      "from %s import %s\nextra_config_dict = %r\n"
      % (RunPromise.__module__, RunPromise.__name__, kw))

  def test_locked_ok(self):
    self.writeLog(self.rf_info_data)
    self.writePromise(sdr_dev='3', sfp_port='1')
    self.configureLauncher()
    self.launcher.run()

  def test_no_lock(self):
    self.writeLog(self.rf_info_data)
    self.writePromise(sdr_dev='2', sfp_port='1')
    self.configureLauncher()
    with self.assertRaisesRegex(PromiseError, r'(?m)HW Lock is missing\n.*SW Lock is missing'):
      self.launcher.run()

  def test_no_device(self):
    self.writeLog(self.rf_info_data)
    self.writePromise(sdr_dev='1', sfp_port='0')
    self.configureLauncher()
    with self.assertRaisesRegex(PromiseError, 'no device entry'):
      self.launcher.run()

  def test_no_cpri_entry(self):
    self.writeLog(self.rf_info_data)
    self.writePromise(sdr_dev='4', sfp_port='0')
    self.configureLauncher()
    with self.assertRaisesRegex(PromiseError, 'no CPRI feature'):
      self.launcher.run()

  def test_stale_data(self):
    self.writeLog(self.rf_info_data, ago=500)
    self.writePromise(sdr_dev='3', sfp_port='1')
    self.configureLauncher()
    with self.assertRaisesRegex(PromiseError, 'rf_info: stale data'):
      self.launcher.run()

  def test_2022(self):
    # Amarisoft software from 2022 has different format for rf_info
    rf_info_data = self.rf_info_data.copy()
    rf_info_data['rf_info'] = \
"""
TRX SDR driver 2022-10-26, API v15/18
PCIe CPRI /dev/sdr0@0:  
  Hardware ID: 0x4b12
  DNA: [0x000841c20971b05c]
  Serial: ''
  FPGA revision: 2022-10-25  08:58:42
  FPGA vccint: 0.98 V
  FPGA vccaux: 1.77 V
  FPGA vccbram: 0.99 V
  FPGA temperature: 71.9 °C
  Clock tune: -0.1 ppm
  NUMA: 0
  CPRI: x16 HW SW
  DMA0: TX fifo: 66.67us  Usage=16/65536 (0%)
  DMA0: RX fifo: 66.67us  Usage=864/65536 (1%)
  DMA0 Underflows: 0
  DMA0 Overflows: 0
  """   # NOTE extra spaces in the end and trailing spaces after /dev/sdr0@:0
    self.writeLog(rf_info_data)
    self.writePromise(sdr_dev='0', sfp_port='0')
    self.configureLauncher()
    self.launcher.run()

  def test_2024_06_15(self):
    # Amarisoft software from 2024-06-15 has changed format of rf_info
    rf_info_data = self.rf_info_data.copy()
    rf_info_data['rf_info'] = \
"""
TRX SDR driver 2024-06-11, API v15
PCIe CPRI /dev/sdr0@0:
  Hardware ID: 0x4b12
  DNA: [0x0068102442f1b05c]
  Serial: ''
  FPGA revision: 2024-03-14  16:19:27
  FPGA vccint: 0.98 V
  FPGA vccaux: 1.77 V
  FPGA vccbram: 0.99 V
  FPGA temperature: 57.6 °C
Clock tune: 0.2 ppm
  NUMA: 0
  CPRI_option: '5' (x8) signal=yes lock=HW+SW rx/tx=2.433us
    Port #0: T14=2.433us
  DMA0: TX fifo: 66.67us  Usage=16/32768 (0%)
  DMA0: RX fifo: 66.67us  Usage=16/32768 (0%)
  DMA0 Underflows: 0
  DMA0: Overflows: 0
  BUFS: TX idx=53800.99 RX idx=53800.100
PCIe CPRI /dev/sdr1@0:
  Hardware ID: 0x4b12
  DNA: [0x0048b504006af054]
  Serial: ''
  FPGA revision: 2024-03-14  16:19:27
  FPGA vccint: 0.99 V
  FPGA vccaux: 1.78 V
  FPGA vccbram: 0.99 V
  FPGA temperature: 51.6 °C
Clock tune: -4.0 ppm
  NUMA: 1
  CPRI_option: '5' (x8) signal=yes lock=HW+SW rx/tx=1.994us
    Port #0: T14=1.994us
  DMA0: TX fifo: 66.67us  Usage=16/32768 (0%)
  DMA0: RX fifo: 66.67us  Usage=16/32768 (0%)
  DMA0 Underflows: 0
  DMA0: Overflows: 0
  BUFS: TX idx=53800.114 RX idx=53800.115
PCIe CPRI /dev/sdr0@2:
  Hardware ID: 0x4b12
  DNA: [0x0068102442f1b05c]
  Serial: ''
  FPGA revision: 2024-03-14  16:19:27
  FPGA vccint: 0.98 V
  FPGA vccaux: 1.77 V
  FPGA vccbram: 0.98 V
  FPGA temperature: 57.6 °C
Clock tune: 0.2 ppm
  NUMA: 0
  CPRI_option: '5' (x8) signal=yes lock=HW+SW rx/tx=2.393us
    Port #0: T14=2.393us
  DMA0: TX fifo: 66.67us  Usage=16/32768 (0%)
  DMA0: RX fifo: 66.67us  Usage=16/32768 (0%)
  DMA0 Underflows: 0
  DMA0: Overflows: 0
  BUFS: TX idx=53800.133 RX idx=53800.134

GPS info:
  UTC:  2024-10-24 11:51:43
  pos:  lat=12.34567°  long=34.56789°
  height: 118.1m  nb_sats: 10
"""
    self.writeLog(rf_info_data)
    for sdr, port in [(0,0), (1,0), (0,2)]:
        self.writePromise(sdr_dev='%d' % sdr, sfp_port='%d' % port)
    self.configureLauncher()
    self.launcher.run()


if __name__ == '__main__':
  unittest.main()



"""
Some notable parameters:

  url:
    The URL of the websocket to test
  promise-timeout:
    Optional timeout (in seconds) for promise.
  timeout:
    Optional timeout (in seconds) for websocket request.
  frequency:
    Optional frequency (in minutes) for running this promise.
  binary:
    Boolean to say if the frames sent to websocket are binary (default) or text, only useful when content* options are set
  content-to-send:
    Optional bytes array or string (depending on binary) to send to the websocket
  content-to-receive:
    Optional bytes array or string (depending on binary) to compare the first message sent by websocket with (must be used with content to send)
  username:
    Optional string containing the user name for basic auth
  password:
    Optional string containing the password for basic auth
"""

from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise
from base64 import b64encode

import ssl
import websocket

def basic_auth(username, password):
  token = b64encode(f"{username}:{password}".encode('utf-8')).decode("ascii")
  return f'Basic {token}'

@implementer(interface.IPromise)
class RunPromise(GenericPromise):
  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # SR can set custom periodicity
    self.setPeriodicity(float(self.getConfig('frequency', 2)))

  def sense(self):
    """
      Check if websocket URL is available.
    """

    url = self.getConfig('url')
    # make default time a max of 5 seconds, a bit smaller than promise-timeout
    # and in the same time at least 1 second
    default_timeout = max(
      1, min(5, int(self.getConfig('promise-timeout', 20)) - 1))
    binary = self.getConfig('binary', True)
    content_to_send = self.getConfig('content-to-send')
    content_to_receive = self.getConfig('content-to-receive')
    username = self.getConfig('username', '')
    password = self.getConfig('password', '')
    if username:
      header = { 'Authorization' : basic_auth(username, password) }
    else:
      header = {}

    try:
      ws = websocket.create_connection(
        url,
        timeout=int(self.getConfig('timeout', default_timeout)),
        header=header,
        sslopt={"cert_reqs": ssl.CERT_NONE}
        )
    except websocket._exceptions.WebSocketBadStatusException:
      self.logger.error(
        "ERROR connection not possible while accessing %r", url)
    except Exception as e:
      self.logger.error("ERROR: %s", e)
    else:
      if content_to_send and content_to_receive:
        if binary:
          ws.send_binary(content_to_send)
        else:
          ws.send(content_to_send)

        response = ws.recv()
        if response != content_to_receive:
          self.logger.error("ERROR received %r instead of %r", response, content_to_receive)
        else:
          self.logger.info("Correctly received %r from %r", content_to_receive, url)
      else:
        self.logger.info("Correctly connected to %r", url)
      ws.close()

  def anomaly(self):
    return self._anomaly(result_count=3, failure_amount=3)



from click.testing import CliRunner
import hashlib
import os
import pathlib
import shutil
import tempfile
import time
import unittest

import slapos.backup_signature


def sorted_listdir(d):
  return sorted(os.listdir(d))


def invoke(argument):
  return CliRunner().invoke(
    slapos.backup_signature.cli,
    argument.split()
  )


class Test(unittest.TestCase):
  def setUp(self):
    self.workdir = tempfile.mkdtemp()
    self.backupdir = os.path.join(self.workdir, 'backup')
    os.mkdir(self.backupdir)

    self.test_data_00 = 'Some test data'
    self.test_data_00_sum = hashlib.sha256(
      self.test_data_00.encode()).hexdigest()
    self.test_file_00 = 'test file'
    with open(os.path.join(self.backupdir, self.test_file_00), 'w') as fh:
     fh.write(self.test_data_00)

    self.test_data_01 = 'Other test data'
    self.test_data_01_sum = hashlib.sha256(
      self.test_data_01.encode()).hexdigest()
    self.test_file_01 = 'test other file'

  def tearDown(self):
    shutil.rmtree(self.workdir)

  def test_update(self):
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {self.backupdir}/backup-signature\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n', fh.read())

  def test_update_directory_with_slash(self):
    self.assertFalse(self.backupdir.endswith('/'))
    result = invoke(
      f'--action update --directory {self.backupdir}// '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {self.backupdir}/backup-signature\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n', fh.read())

  def test_update_bad_line(self):
    with open(os.path.join(self.backupdir, 'backup-signature'), 'w') as fh:
     fh.write('badline\n')
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      "WARNING: Bad line badline\n"
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {self.backupdir}/backup-signature\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n', fh.read())

  def test_update_create_proof_signature(self):
    proof_signature = os.path.join(self.workdir, 'proof.signature')
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature '
      f'--proof-signature-path {proof_signature}')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {proof_signature}\n"
    )
    self.assertEqual(
      [f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(proof_signature) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n', fh.read())

  def test_update_deep_signature(self):
    os.mkdir(os.path.join(self.backupdir, 'path'))
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name path/backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {self.backupdir}/path/backup-signature\n"
    )
    self.assertEqual(
      ['path', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'path', 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n', fh.read())

  def test_update_noop(self):
    self.test_update()
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Kept ./{self.test_file_00}\n"
      f"INFO: Kept {self.backupdir}/backup-signature\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n', fh.read())

  def test_update_proof_signature(self):
    self.test_update()
    proof_signature = os.path.join(self.workdir, 'proof.signature')
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature '
      f'--proof-signature-path {proof_signature}')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {proof_signature}\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n', fh.read())

    with open(proof_signature) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n', fh.read())

  def test_update_noop_force(self):
    self.test_update()
    result = invoke(
      f'--action update --force --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {self.backupdir}/backup-signature\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n', fh.read())

  def test_update_change(self):
    self.test_update()
    time.sleep(.5)
    with open(os.path.join(self.backupdir, self.test_file_00), 'w') as fh:
     fh.write(self.test_data_01)
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {self.backupdir}/backup-signature\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_01_sum}  ./{self.test_file_00}\n', fh.read())

  def test_update_change_proof_signature(self):
    self.test_update()
    proof_signature = os.path.join(self.workdir, 'proof.signature')
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature '
      f'--proof-signature-path {proof_signature}')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {proof_signature}\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(proof_signature) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n', fh.read())
    time.sleep(.5)
    with open(os.path.join(self.backupdir, self.test_file_00), 'w') as fh:
     fh.write(self.test_data_01)
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {self.backupdir}/backup-signature\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_01_sum}  ./{self.test_file_00}\n', fh.read())

    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature '
      f'--proof-signature-path {proof_signature}')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {proof_signature}\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(proof_signature) as fh:
      self.assertEqual(
        f'{self.test_data_01_sum}  ./{self.test_file_00}\n', fh.read())

  def test_update_change_older(self):
    self.test_update()
    time.sleep(.5)
    with open(os.path.join(self.backupdir, self.test_file_00), 'w') as fh:
     fh.write(self.test_data_01)
    time.sleep(.5)
    pathlib.Path(os.path.join(self.backupdir, 'backup-signature')).touch()
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Kept ./{self.test_file_00}\n"
      f"INFO: Kept {self.backupdir}/backup-signature\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n', fh.read())

    # force is needed
    result = invoke(
      f'--action update --force --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Updated ./{self.test_file_00}\n"
      f"INFO: Updated {self.backupdir}/backup-signature\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_01_sum}  ./{self.test_file_00}\n', fh.read())

  def test_update_add(self):
    self.test_update()
    with open(os.path.join(self.backupdir, self.test_file_01), 'w') as fh:
     fh.write(self.test_data_01)
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Kept ./{self.test_file_00}\n"
      f"INFO: Updated ./{self.test_file_01}\n"
      f"INFO: Updated {self.backupdir}/backup-signature\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}', f'{self.test_file_01}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n'
        f'{self.test_data_01_sum}  ./{self.test_file_01}\n',
        fh.read())

  def test_update_remove(self):
    self.test_update_add()
    os.unlink(os.path.join(self.backupdir, self.test_file_01))
    result = invoke(
      f'--action update --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"INFO: Kept ./{self.test_file_00}\n"
      f"INFO: Updated {self.backupdir}/backup-signature\n"
    )
    self.assertEqual(
      ['backup-signature', f'{self.test_file_00}'],
      sorted_listdir(self.backupdir)
    )
    with open(os.path.join(self.backupdir, 'backup-signature')) as fh:
      self.assertEqual(
        f'{self.test_data_00_sum}  ./{self.test_file_00}\n',
        fh.read())

  def test_validate(self):
    backup_signature = os.path.join(self.backupdir, 'backup-signature')
    with open(backup_signature, 'w') as fh:
     fh.write(f'{self.test_data_00_sum}  ./{self.test_file_00}\n')
    result = invoke(
      f'--action validate --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"DEBUG: Calculated ./{self.test_file_00}\n"
      "OK: Signature match.\n"
    )

  def test_validate_bad_line(self):
    backup_signature = os.path.join(self.backupdir, 'backup-signature')
    with open(backup_signature, 'w') as fh:
     fh.write(
       'badline\n'
       f'{self.test_data_00_sum}  ./{self.test_file_00}\n'
     )
    result = invoke(
      f'--action validate --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 2)
    self.assertEqual(
      result.output,
      "ERROR: Bad line badline\n"
      f"DEBUG: Calculated ./{self.test_file_00}\n"
      "OK: Signature match.\n"
      f"ERROR: Signature {backup_signature} errors: 1\n"
    )

  def test_validate_no_match(self):
    backup_signature = os.path.join(self.backupdir, 'backup-signature')
    with open(backup_signature, 'w') as fh:
     fh.write(f'{self.test_data_00_sum}a  ./{self.test_file_00}\n')
    result = invoke(
      f'--action validate --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 1)
    self.assertEqual(
      result.output,
      f"DEBUG: Calculated ./{self.test_file_00}\n"
      "ERROR: Signatures do not match, current signature:\n"
      f"{{'./{self.test_file_00}': '{self.test_data_00_sum}a'}}\n"
      "Calculated signature:\n"
      f"{{'./{self.test_file_00}': '{self.test_data_00_sum}'}}\n"
    )

  def test_validate_missing(self):
    result = invoke(
      f'--action validate --directory {self.backupdir} '
      '--signature-file-name backup-signature')
    self.assertEqual(result.exit_code, 1)
    self.assertEqual(
      result.output,
      f"ERROR: Signature file {self.backupdir}/backup-signature not found\n"
    )

  def test_validate_timestamp(self):
    backup_signature = os.path.join(self.backupdir, 'backup-signature')
    validate_timestamp = os.path.join(self.workdir, 'validate-timestamp')
    with open(backup_signature, 'w') as fh:
     fh.write(f'{self.test_data_00_sum}  ./{self.test_file_00}\n')
    time.sleep(0.5)
    result = invoke(
      f'--action validate --directory {self.backupdir} '
      '--signature-file-name backup-signature '
      f'--validate-timestamp-file {validate_timestamp}')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"DEBUG: Calculated ./{self.test_file_00}\n"
      "OK: Signature match.\n"
      f"DEBUG: Updated {validate_timestamp}\n"
    )

    result = invoke(
      f'--action validate --directory {self.backupdir} '
      '--signature-file-name backup-signature '
      f'--validate-timestamp-file {validate_timestamp}')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"DEBUG: Using {validate_timestamp}\n"
      f"DEBUG: Skipped ./{self.test_file_00}\n"
      "OK: Signature match.\n"
      f"DEBUG: Updated {validate_timestamp}\n"
    )

  def test_validate_timestamp_update(self):
    backup_signature = os.path.join(self.backupdir, 'backup-signature')
    validate_timestamp = os.path.join(self.workdir, 'validate-timestamp')
    self.test_validate_timestamp()
    time.sleep(0.5)
    with open(os.path.join(self.backupdir, self.test_file_01), 'w') as fh:
     fh.write(self.test_data_01)
    with open(backup_signature, 'w') as fh:
     fh.write(
       f'{self.test_data_00_sum}  ./{self.test_file_00}\n'
       f'{self.test_data_01_sum}  ./{self.test_file_01}\n'
     )
    result = invoke(
      f'--action validate --directory {self.backupdir} '
      '--signature-file-name backup-signature '
      f'--validate-timestamp-file {validate_timestamp}')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"DEBUG: Using {validate_timestamp}\n"
      f"DEBUG: Skipped ./{self.test_file_00}\n"
      f"DEBUG: Calculated ./{self.test_file_01}\n"
      "OK: Signature match.\n"
      f"DEBUG: Updated {validate_timestamp}\n"
    )

  def test_validate_timestamp_update_force(self):
    self.test_validate_timestamp_update()
    validate_timestamp = os.path.join(self.workdir, 'validate-timestamp')
    result = invoke(
      f'--action validate --directory {self.backupdir} '
      '--force '
      '--signature-file-name backup-signature '
      f'--validate-timestamp-file {validate_timestamp}')
    self.assertEqual(result.exit_code, 0)
    self.assertEqual(
      result.output,
      f"DEBUG: Calculated ./{self.test_file_00}\n"
      f"DEBUG: Calculated ./{self.test_file_01}\n"
      "OK: Signature match.\n"
      f"DEBUG: Updated {validate_timestamp}\n"
    )



# coding: utf-8
# Copyright (C) 2025  Nexedi SA and Contributors.
#                     Łukasz Nowak <luke@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import click
import hashlib
import os
import pprint
import sys
import time


def update(force, directory, signature_file_name, proof_signature_path):
  if proof_signature_path is not None:
    signature_file = proof_signature_path
  else:
    signature_file = os.path.join(directory, signature_file_name)
  current_signature = {}
  new_signature = {}
  if force or not os.path.exists(signature_file):
    signature_modification = 0
  else:
    signature_modification = os.path.getmtime(signature_file)
    with open(signature_file) as fh:
      for line in fh.readlines():
        line = line.strip()
        line_split = line.split(maxsplit=1)
        if len(line_split) != 2:
          print(f'WARNING: Bad line {line}')
          continue
        checksum = line_split[0]
        filename = line_split[1]
        current_signature[filename] = checksum
  for root, directory_list, file_list in os.walk(directory):
    for filename in sorted(file_list):
      filepath = os.path.join(root, filename)
      signature_path = filepath.replace(directory, './')
      if signature_path == './' + signature_file_name:
        continue
      file_modification = os.path.getmtime(filepath)
      if file_modification > signature_modification \
         or signature_path not in current_signature:
        with open(filepath, 'rb') as fh:
          file_hash = hashlib.sha256()
          while chunk := fh.read(2**20):
            file_hash.update(chunk)
          new_signature[signature_path] = file_hash.hexdigest()
        print(f'INFO: Updated {signature_path}')
      else:
        print(f'INFO: Kept {signature_path}')
        new_signature[signature_path] = current_signature[signature_path]

  if new_signature != current_signature or not os.path.exists(signature_file):
    with open(signature_file, 'w') as fh:
      for signature_path in sorted(new_signature):
        checksum = new_signature[signature_path]
        fh.write(f'{checksum}  {signature_path}\n')
    print(f'INFO: Updated {signature_file}')
  else:
    print(f'INFO: Kept {signature_file}')


def validate(force, directory, signature_file_name, validate_timestamp_file):
  signature_file = os.path.join(directory, signature_file_name)
  if not os.path.exists(signature_file):
    print(f'ERROR: Signature file {signature_file} not found')
    sys.exit(1)

  if force or validate_timestamp_file is None \
     or not os.path.exists(validate_timestamp_file):
    validate_timestamp = 0
  else:
    print(f'DEBUG: Using {validate_timestamp_file}')
    validate_timestamp = os.path.getmtime(validate_timestamp_file)

  current_signature = {}
  new_signature = {}
  signature_error_count = 0
  with open(signature_file) as fh:
    for line in fh.readlines():
      line = line.strip()
      line_split = line.split(maxsplit=1)
      if len(line_split) != 2:
        print(f'ERROR: Bad line {line}')
        signature_error_count += 1
        continue
      checksum = line_split[0]
      filename = line_split[1]
      current_signature[filename] = checksum

  for root, directory_list, file_list in os.walk(directory):
    for filename in sorted(file_list):
      filepath = os.path.join(root, filename)
      signature_path = filepath.replace(directory, './')
      if signature_path == './' + signature_file_name:
        continue
      file_modification = os.path.getmtime(filepath)
      if signature_path in current_signature \
         and file_modification < validate_timestamp:
        new_signature[signature_path] = current_signature[signature_path]
        print(f'DEBUG: Skipped {signature_path}')
      else:
        with open(filepath, 'rb') as fh:
          file_hash = hashlib.sha256()
          while chunk := fh.read(2**20):
            file_hash.update(chunk)
          new_signature[signature_path] = file_hash.hexdigest()
        print(f'DEBUG: Calculated {signature_path}')

  if new_signature != current_signature:
    print('ERROR: Signatures do not match, current signature:')
    pprint.pprint(current_signature)
    print('Calculated signature:')
    pprint.pprint(new_signature)
    sys.exit(1)
  else:
    print('OK: Signature match.')
    if validate_timestamp_file is not None:
      with open(validate_timestamp_file, 'w') as fh:
        fh.write(str(time.time()))
      print(f'DEBUG: Updated {validate_timestamp_file}')
    if signature_error_count > 0:
      print(
        f'ERROR: Signature {signature_file} errors: {signature_error_count}')
      sys.exit(2)


@click.command(short_help="Backup signature handling")
@click.option(
  '--action',
  type=click.Choice(['update', 'validate'], case_sensitive=False),
  required=True,
  help="Action to take"
)
@click.option(
  '--directory',
  type=click.Path(),
  required=True,
  help="Directory to work in"
)
@click.option(
  '--signature-file-name',
  type=click.Path(),
  required=True,
  help="Name of signature file, expected in top of the --directory"
)
@click.option(
  '--proof-signature-path',
  type=click.Path(),
  help="Path to proof signature, which will be updated instead of "
       "backup signature"
)
@click.option(
  '--validate-timestamp-file',
  type=click.Path(),
  help="Location of file which modification time relates to last validation, "
       "so that only files not validated before will be checksummed"
)
@click.option(
  '--force',
  is_flag=True,
  default=False,
  show_default=True,
  help="Forces full run"
)
def cli(
  action, force, directory, signature_file_name, validate_timestamp_file,
  proof_signature_path):
  """
  Tool to handle backup and validation signatures

  The most useful action is update, which will update the signature-file-name
  inside of the directory. When proof-signature-path is provided, it'll fill
  it of checksums from directory, ignoring signature-file-name.

  The validate action allows to check is files are matching the checksums
  stored in signature-file-name. When validate-timestamp-file is provided this
  file will be used to only check newly appeared files.

  --force option allows to do the check or validate without any optimisation.
  """
  directory = directory.rstrip('/') + '/'
  if action == 'update':
    update(force, directory, signature_file_name, proof_signature_path)
  elif action == 'validate':
    validate(force, directory, signature_file_name, validate_timestamp_file)



from .util import get_json_log_data_interval
from .util import JSONPromise

import json
from zope.interface import implementer
from slapos.grid.promise import interface

@implementer(interface.IPromise)
class RunPromise(JSONPromise):

  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    self.setPeriodicity(float(self.getConfig('frequency', 1)))
    self.testing = self.getConfig('testing') == "True"
    self.amarisoft_stats_log = self.getConfig('amarisoft-stats-log')
    self.stats_period = int(self.getConfig('stats-period'))
    self.rx_chan_list = json.loads(self.getConfig('rf-rx-chan-list')) # which rx channels to check
    self.max_rx_sample_db = float(self.getConfig('max-rx-sample-db'))

  def sense(self):

    data_list = get_json_log_data_interval(self.amarisoft_stats_log, self.stats_period * 2)

    max_rx_list = []
    saturated = False

    for rx_antenna_list in map(lambda x: x['samples']['rx'], data_list):
        rx_list = map(lambda x: float(x['max']), [rx_antenna_list[i] for i in self.rx_chan_list])
        if not max_rx_list:
          max_rx_list = list(rx_list)
        for i, rx in enumerate(rx_list):
            max_rx_list[i] = max(max_rx_list[i], rx)
            if rx >= self.max_rx_sample_db:
                saturated = True

    self.json_logger.info("RX maximum sample values (dB)", extra={'data': max_rx_list})

    if not max_rx_list:
        self.logger.error("No RX samples data available")
    elif saturated:
        self.logger.error("RX antennas saturated, please lower rx_gain")
    else:
        self.logger.info("No saturation detected on RX antennas")

  def test(self):
    """
      Called after sense() if the instance is still converging.
      Returns success or failure based on sense results.

      In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)

  def anomaly(self):
    """
      Called after sense() if the instance has finished converging.
      Returns success or failure based on sense results.
      Failure signals the instance has diverged.

      In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=1, failure_amount=1)



import json
import os
import psutil
import time

from psutil._common import bytes2human
from .util import JSONPromise

from zope.interface import implementer
from slapos.grid.promise import interface

@implementer(interface.IPromise)
class RunPromise(JSONPromise):

  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # Get reference values
    self.setPeriodicity(float(self.getConfig('frequency', 2)))
    self.last_avg_ram_file = self.getConfig('last-avg-ram-file', 'last_avg')
    self.min_threshold_ram = float(self.getConfig('min-threshold-ram', 500))*1048576 #  MB converted into bytes
    self.min_avg_ram = float(self.getConfig('min-avg-ram', 1e3))*1048576 #  MB converted into bytes
    self.avg_ram_period =  int(self.getConfig('avg-ram-period', 600)) # secondes

  def sense(self):
    promise_success = True
    # Get current RAM usage
    ram_data = psutil.virtual_memory()
    # Check with min threshold and log error if below it
    if ram_data.available <= self.min_threshold_ram:
      self.logger.error("RAM usage reached critical threshold: %7s "\
        " (threshold is %7s)" % (bytes2human(ram_data.available), bytes2human(self.min_threshold_ram)))
      promise_success = False
    
    # Log RAM usage
    data = json.dumps({'available_ram': ram_data.available})
    self.json_logger.info("RAM data", extra={'data': data})

    # Get last timestamp (i.e. last modification) of log file
    try:
      t = os.path.getmtime(self.last_avg_ram_file)
    except OSError:
      t = 0
    # Get last available RAM from log file since avg_ram_period / 4
    if (time.time() - t) > self.avg_ram_period / 4:
      open(self.last_avg_ram_file, 'w').close()
      temp_list = self.get_json_log_data_interval(self.avg_ram_period)
      if temp_list:
        avg_ram = sum(map(lambda x: x['available_ram'], temp_list)) / len(temp_list)
        if avg_ram < self.min_avg_ram:
          self.logger.error("Average RAM usage over the last %s seconds "\
            "reached threshold: %7s (threshold is %7s)" 
            % (self.avg_ram_period, bytes2human(avg_ram), bytes2human(self.min_avg_ram)))
          promise_success = False
      else:
        self.logger.error("Couldn't read available RAM from log")
        promise_success = False

    if promise_success:
      self.logger.info("RAM usage OK")

  def test(self):
    """
    Called after sense() if the instance is still converging.
    Returns success or failure based on sense results.

    In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)


  def anomaly(self):
    """
    Called after sense() if the instance has finished converging.
    Returns success or failure based on sense results.
    Failure signals the instance has diverged.

    In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=3, failure_amount=2)



##############################################################################
#
# Copyright (c) 2019 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

# This module contains python3 syntax that can't be parsed by python2
# that's why it is in a separated module

from slapos.grid.promise import PromiseError
from . import TestPromisePluginMixin
from slapos.util import str2bytes

import asyncio
import contextlib
import os
import time
import websocket
from websockets import serve
import multiprocessing


class CheckWebsocketAvailableMixin(TestPromisePluginMixin):


  @classmethod
  def setUpClass(cls):
    SLAPOS_TEST_IPV4 = os.environ.get('SLAPOS_TEST_IPV4', '127.0.0.1')
    SLAPOS_TEST_IPV4_PORT = 57965
    cls.WS_ENDPOINT = "ws://%s:%s/" % (SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV4_PORT)

    async def echo(websocket):
      path = websocket.path.split('/')[-1]

      if '_' in path:
        response, timeout = path.split('_')
        response = response
        timeout = int(timeout)
      else:
        timeout = 0
        response = path

      time.sleep(timeout)
      async for message in websocket:
        if response == "OK":
          await websocket.send(message)
        else:
          await websocket.send("bad")

    async def server():
      async with serve(echo, SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV4_PORT):
        await asyncio.Future()  # run forever

    def main():
      asyncio.run(server())

    cls.server_process = multiprocessing.Process(target=main)
    cls.server_process.start()
    for _ in range(20):
      try:
        with contextlib.closing(websocket.create_connection((SLAPOS_TEST_IPV4, SLAPOS_TEST_IPV4_PORT))):
          break
      except Exception:
        time.sleep(.1)

  @classmethod
  def tearDownClass(cls):
    cls.server_process.terminate()
    cls.server_process.join()

  def setUp(self):
    TestPromisePluginMixin.setUp(self)
    self.promise_name = "check-websocket-available.py"

  def make_content(self, option_dict):
    content = """from slapos.promise.plugin.check_websocket_available import RunPromise

extra_config_dict = {
"""
    for option in option_dict.items():
      content += "\n  '%s': %r," % option

    return content + "\n}"

  def tearDown(self):
    TestPromisePluginMixin.tearDown(self)



from .util import get_json_log_data_interval
from .util import JSONPromise

from zope.interface import implementer
from slapos.grid.promise import interface

@implementer(interface.IPromise)
class RunPromise(JSONPromise):
  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    self.setPeriodicity(float(self.getConfig('frequency', 1)))
    self.amarisoft_stats_log = self.getConfig('amarisoft-stats-log')
    self.stats_period = int(self.getConfig('stats-period'))
    self.min_rxtx_delay_threshold = float(self.getConfig('min-rxtx-delay', 0))
    self.testing = self.getConfig('testing') == "True"

  def sense(self):

    if self.testing:
        self.logger.info("skipping promise")
        return

    data_list = get_json_log_data_interval(self.amarisoft_stats_log, self.stats_period * 5)

    min_rxtx_delay_it = map(lambda x: float(x['rf']['rxtx_delay_min']), data_list)
    if not min_rxtx_delay_it:
        self.logger.error("No TX/RX diff data available")
    else:
      min_rxtx_delay = min(min_rxtx_delay_it)
      if min_rxtx_delay < self.min_rxtx_delay_threshold:
        self.logger.error("The minimum available time for radio front end processing is lower than the minimum threshold (%s ms)." % (self.min_rxtx_delay_threshold,))
      else:
        self.logger.info("The minimum available time for radio front end processing is higher than the minimum threshold (%s ms)." % (self.min_rxtx_delay_threshold,))

    self.json_logger.info("Min RX TX Delay (ms)",
      extra={'data': {'min_rxtx_delay': min_rxtx_delay}})

  def test(self):
    """
      Called after sense() if the instance is still converging.
      Returns success or failure based on sense results.

      In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)

  def anomaly(self):
    """
      Called after sense() if the instance has finished converging.
      Returns success or failure based on sense results.
      Failure signals the instance has diverged.

      In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=1, failure_amount=1)



##############################################################################
#
# Copyright (c) 2019 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################



from slapos.grid.promise import PromiseError
from slapos.util import str2bytes

import unittest
import sys
import contextlib
import os
import random
import string
import time
import multiprocessing

if sys.version_info[0] >= 3:
  import asyncio
  import websocket
  from websockets import serve
  from .check_websocket_available_py3_test import CheckWebsocketAvailableMixin
else:
  class CheckWebsocketAvailableMixin():
    pass

@unittest.skipIf(sys.version_info[0] < 3, "not supported in this library version")
class TestCheckWebsocketAvailable(CheckWebsocketAvailableMixin):

  def test_check_url_bad(self):
    content = self.make_content({
      'url': 'ws://',
      'timeout': 10,
    })
    self.writePromise(self.promise_name, content)
    self.configureLauncher()
    with self.assertRaises(PromiseError):
      self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    self.assertEqual(result['result']['failed'], True)
    self.assertEqual(
      result['result']['message'],
      "ERROR: hostname is invalid"
    )

  def test_check_simple_connect(self):
    url = self.WS_ENDPOINT + 'OK'
    content = self.make_content({
      'url': url,
      'timeout': 10,
    })
    self.writePromise(self.promise_name, content)
    self.configureLauncher()
    self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    self.assertEqual(result['result']['failed'], False)
    self.assertEqual(
      result['result']['message'],
      ("Correctly connected to %r" % url)
    )

  def test_check_read_text(self):
    text = ''.join(random.choice(string.ascii_letters) for i in range(10))
    url = self.WS_ENDPOINT + 'OK'
    content = self.make_content({
      'url': url,
      'binary': False,
      'content-to-send': text,
      'content-to-receive' : text
    })
    self.writePromise(self.promise_name, content)
    self.configureLauncher()
    self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    self.assertEqual(result['result']['failed'], False)
    self.assertEqual(
      result['result']['message'],
      ("Correctly received %r from %r" % (text, url))
    )

  def test_check_read_binary(self):
    text = os.urandom(100)
    url = self.WS_ENDPOINT + 'OK'
    content = self.make_content({
      'url': url,
      'content-to-send': text,
      'content-to-receive' : text
    })
    self.writePromise(self.promise_name, content)
    self.configureLauncher()
    self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    self.assertEqual(result['result']['failed'], False)
    self.assertEqual(
      result['result']['message'],
      ("Correctly received %r from %r" % (text, url))
    )

  def test_check_bad_read(self):
    text = ''.join(random.choice(string.ascii_letters) for i in range(10))
    url = self.WS_ENDPOINT + 'NOK'
    content = self.make_content({
      'url': url,
      'content-to-send': text,
      'content-to-receive' : text
    })
    self.writePromise(self.promise_name, content)
    self.configureLauncher()
    with self.assertRaises(PromiseError):
      self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    self.assertEqual(result['result']['failed'], True)
    self.assertEqual(
      result['result']['message'],
      ("ERROR received 'bad' instead of %r" % text)
    )

  def test_check_timeout(self):
    url = self.WS_ENDPOINT + 'OK_5'
    content = self.make_content({
      'url': url,
      'timeout': 1,
      # use content to send/receceive so that promise will try to read from websocket
      # otherwise, we can't test the timeout
      'content-to-send': "a",
      'content-to-receive' : "a",
    })
    self.writePromise(self.promise_name, content)
    self.configureLauncher()
    with self.assertRaises(PromiseError):
      self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    self.assertEqual(result['result']['failed'], True)
    self.assertEqual(
      result['result']['message'],
      "Error: Promise timed out after 0.5 seconds",
    )

if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
##############################################################################
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
##############################################################################

import mock
import time
from slapos.grid.promise import PromiseError
from slapos.promise.plugin.check_cpu_temperature import RunPromise
from . import TestPromisePluginMixin


class TestCheckCpuTemperature(TestPromisePluginMixin):

  promise_name = "monitor-cpu-temperature.py"

  def setUp(self):
    super(TestCheckCpuTemperature, self).setUp()

  def writePromise(self, **kw):
    super(TestCheckCpuTemperature, self).writePromise(self.promise_name,
      "from %s import %s\nextra_config_dict = %r\n"
      % (RunPromise.__module__, RunPromise.__name__, kw))

  def runPromise(self, summary, failed=False):
    self.configureLauncher(enable_anomaly=True, force=True)
    with mock.patch('psutil.sensors_temperatures', return_value=summary):
      if failed:
        self.assertRaises(PromiseError, self.launcher.run)
      else:
        self.launcher.run()
    result = self.getPromiseResult(self.promise_name)['result']
    self.assertEqual(result['failed'], failed)
    return result['message']

  def test_temp_ok(self):
    message = "Temperature OK (50 °C)"
    self.writePromise(**{
        'last-avg-computation-file':'last_avg_computation_file',
        'max-spot-temp': 80,
        'max-avg-temp': 100,
    })
    self.assertEqual(message, self.runPromise({'coretemp': [[0, 50]]}))

  def test_spot_critical(self):
    message = "Temperature reached critical threshold: 90 °C (threshold is 80.0 °C)"
    self.writePromise(**{
        'last-avg-computation-file':'last_avg_computation_file',
        'max-spot-temp': 80,
        'max-avg-temp': 100,
    })
    self.assertEqual(message, self.runPromise({'coretemp': [[0, 90]]}))

  def test_avg_critical(self):
    message = "Average temperature over the last 1s reached threshold: 45.0 °C (threshold is 40.0 °C)"
    self.writePromise(**{
        'last-avg-computation-file':'last_avg_computation_file',
        'max-spot-temp': 99999,
        'max-avg-temp': 40,
        'avg-temp-duration': 1,
    })
    m = self.runPromise({'coretemp': [[0, 0]]})
    time.sleep(0.6)
    m = self.runPromise({'coretemp': [[0, 0]]})
    time.sleep(0.5)
    self.assertEqual(message, self.runPromise({'coretemp': [[0, 90]]}))

if __name__ == '__main__':
  unittest.main()



from slapos.grid.promise import PromiseError
from slapos.test.promise.plugin import TestPromisePluginMixin

import email
import json
import os
import shutil
import tempfile
import time


class CheckSurykatkaJSONMixin(TestPromisePluginMixin):
  maxDiff = None  # show full diffs for test readability
  promise_name = 'check-surykatka-json.py'

  def setUp(self):
    self.working_directory = tempfile.mkdtemp()
    self.json_file = os.path.join(self.working_directory, 'surykatka.json')
    self.addCleanup(shutil.rmtree, self.working_directory)
    TestPromisePluginMixin.setUp(self)

    now = time.time()
    minute = 60
    day = 24 * 3600
    create_date = email.utils.formatdate
    self.time_past14d = create_date(now - 14 * day)
    self.time_past29d = create_date(now - 29 * day)
    self.time_past20m = create_date(now - 20 * minute)
    self.time_past2m = create_date(now - 2 * minute)
    self.time_future20m = create_date(now + 20 * minute)
    self.time_future3d = create_date(now + 3 * day)
    self.time_future14d = create_date(now + 14 * day)
    self.time_future29d = create_date(now + 29 * day)
    self.time_future60d = create_date(now + 60 * day)

  def writeSurykatkaPromise(self, d=None):
    if d is None:
      d = {}
    content_list = [
      "from slapos.promise.plugin.check_surykatka_json import RunPromise"]
    content_list.append('extra_config_dict = {')
    for k, v in d.items():
      content_list.append("  '%s': '%s'," % (k, v))
    content_list.append('}')
    self.writePromise(self.promise_name, '\n'.join(content_list))

  def writeSurykatkaJsonDirect(self, content):
    with open(self.json_file, 'w') as fh:
      fh.write(content)

  def writeSurykatkaJson(self, content):
    with open(self.json_file, 'w') as fh:
      json.dump(content, fh, indent=2)

  def assertFailedMessage(self, result, message):
    self.assertEqual(result['result']['failed'], True)
    self.assertEqual(
      result['result']['message'],
      message)

  def assertPassedMessage(self, result, message):
    self.assertEqual(result['result']['failed'], False)
    self.assertEqual(
      result['result']['message'],
      message)

  def runAndAssertPassedMessage(self, message):
    self.configureLauncher(enable_anomaly=True)
    self.launcher.run()
    self.assertPassedMessage(
      self.getPromiseResult(self.promise_name),
      message
    )

  def runAndAssertFailedMessage(self, message):
    self.configureLauncher(enable_anomaly=True)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      message
    )


class TestCheckSurykatkaJSONBase(CheckSurykatkaJSONMixin):
  def test_no_config(self):
    self.writeSurykatkaPromise()
    self.configureLauncher(enable_anomaly=True)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "ERROR File '' does not exists")

  def test_not_existing_file(self):
    self.writeSurykatkaPromise({'json-file': self.json_file})
    self.configureLauncher(enable_anomaly=True)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "ERROR File '%s' does not exists" % (self.json_file,))

  def test_empty_file(self):
    self.writeSurykatkaPromise({'json-file': self.json_file})
    self.writeSurykatkaJsonDirect('')
    self.configureLauncher(enable_anomaly=True)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "ERROR loading JSON from '%s'" % (self.json_file,))

  def test(self):
    self.writeSurykatkaPromise(
      {
        'report': 'NOT_EXISTING_ENTRY',
        'json-file': self.json_file,
      }
    )
    self.writeSurykatkaJson({})
    self.configureLauncher(enable_anomaly=True)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "ERROR Report 'NOT_EXISTING_ENTRY' is not supported")


class TestCheckSurykatkaJSONBotStatus(CheckSurykatkaJSONMixin):
  def test(self):
    self.writeSurykatkaPromise(
      {
        'report': 'bot_status',
        'json-file': self.json_file,
      }
    )
    self.writeSurykatkaJson({
      "bot_status": [
        {
          "date": self.time_past2m,
          "text": "loop"}]})
    self.configureLauncher(enable_anomaly=True)
    self.launcher.run()
    self.assertPassedMessage(
      self.getPromiseResult(self.promise_name),
      "bot_status: OK Last bot status"
    )

  def test_no_loop(self):
    self.writeSurykatkaPromise(
      {
        'report': 'bot_status',
        'json-file': self.json_file,
      }
    )
    self.writeSurykatkaJson({
      "bot_status": [
        {
          "date": "Wed, 13 Dec 2222 09:10:11 -0000",
          "text": "error"}]})
    self.configureLauncher(enable_anomaly=True)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "bot_status: ERROR bot_status is 'error' instead of 'loop' in '%s'" % (
        self.json_file,)
    )

  def test_bot_status_future(self):
    self.writeSurykatkaPromise(
      {
        'report': 'bot_status',
        'json-file': self.json_file,
      }
    )
    self.writeSurykatkaJson({
      "bot_status": [
        {
          "date": self.time_future20m,
          "text": "loop"}]})
    self.configureLauncher(enable_anomaly=True)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "bot_status: ERROR Last bot datetime is in future"
    )

  def test_bot_status_old(self):
    self.writeSurykatkaPromise(
      {
        'report': 'bot_status',
        'json-file': self.json_file,
      }
    )
    self.writeSurykatkaJson({
      "bot_status": [
        {
          "date": self.time_past20m,
          "text": "loop"}]})
    self.configureLauncher(enable_anomaly=True)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "bot_status: ERROR Last bot datetime is more than 15 minutes old"
    )

  def test_not_bot_status(self):
    self.writeSurykatkaPromise(
      {
        'report': 'bot_status',
        'json-file': self.json_file,
      }
    )
    self.writeSurykatkaJson({})
    self.configureLauncher(enable_anomaly=True)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "bot_status: ERROR 'bot_status' not in '%s'" % (self.json_file,))

  def test_empty_bot_status(self):
    self.writeSurykatkaPromise(
      {
        'report': 'bot_status',
        'json-file': self.json_file,
      }
    )
    self.writeSurykatkaJson({"bot_status": []})
    self.configureLauncher(enable_anomaly=True)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "bot_status: ERROR 'bot_status' empty in '%s'" % (self.json_file,))


class TestCheckSurykatkaJSONHttpQueryOnlyIPV4(CheckSurykatkaJSONMixin):
  def writeSurykatkaPromise(self, d):
    d.update(**{
        'report': 'http_query',
        'json-file': self.json_file,
    })
    super().writeSurykatkaPromise(d)

  def setUp(self):
    super().setUp()
    self.writeSurykatkaJson({
      "http_query": [
        {
          "ip": "127.0.0.1",
          "status_code": 302,
          "url": "https://www.allok.com/",
          "total_seconds": 4
        },
        {
          "ip": "127.0.0.2",
          "status_code": 302,
          "url": "https://www.allok.com/",
          "total_seconds": 4
        },
        {
          "ip": "127.0.0.1",
          "status_code": 302,
          "url": "http://www.httpallok.com/",
          "total_seconds": 4
        },
        {
          "ip": "127.0.0.2",
          "status_code": 302,
          "url": "http://www.httpallok.com/",
          "total_seconds": 4
        },
      ],
      "ssl_certificate": [
        {
          "hostname": "www.allok.com",
          "ip": "127.0.0.1",
          "not_after": self.time_future60d
        },
        {
          "hostname": "www.allok.com",
          "ip": "127.0.0.2",
          "not_after": self.time_future60d
        },
      ],
      "dns_query": [
        {
            "domain": "www.allok.com",
            "rdtype": "A",
            "resolver_ip": "1.2.3.4",
            "response": "127.0.0.1, 127.0.0.2"
        },
        {
            "domain": "www.httpallok.com",
            "rdtype": "A",
            "resolver_ip": "1.2.3.4",
            "response": "127.0.0.1, 127.0.0.2"
        },
      ],
      "tcp_server": [
        {
            "ip": "127.0.0.1",
            "state": "open",
            "port": 443,
            "domain": "www.allok.com"
        },
        {
            "ip": "127.0.0.2",
            "state": "open",
            "port": 443,
            "domain": "www.allok.com"
        },
        {
            "ip": "127.0.0.1",
            "state": "open",
            "port": 80,
            "domain": "www.httpallok.com"
        },
        {
            "ip": "127.0.0.2",
            "state": "open",
            "port": 80,
            "domain": "www.httpallok.com"
        },
      ],
      "whois": [
        {
            "domain": "allok.com",
            "expiration_date": self.time_future60d,
        },
        {
            "domain": "httpallok.com",
            "expiration_date": self.time_future60d,
        },
      ]
    })

  def test_all_ok(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
        'ip-list': '127.0.0.1 127.0.0.2',
        'maximum-elapsed-time': '5',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "dns_query: OK resolver's 1.2.3.4: 127.0.0.1 127.0.0.2 "
      "whois: OK allok.com expires in > 30 days "
      "tcp_server: OK IP 127.0.0.1:443 OK IP 127.0.0.2:443 "
      "http_query: OK IP 127.0.0.1 status_code 302 OK IP 127.0.0.2 "
      "status_code 302 "
      "ssl_certificate: OK IP 127.0.0.1 expires in > 15 days OK IP "
      "127.0.0.2 expires in > 15 days "
      "elapsed_time: OK IP 127.0.0.1 replied < 5.00s OK IP 127.0.0.2 replied "
      "< 5.00s"
    )

  def test_http_all_ok(self):
    self.writeSurykatkaPromise(
      {
        'report': 'http_query',
        'json-file': self.json_file,
        'url': 'http://www.httpallok.com/',
        'status-code': '302',
        'ip-list': '127.0.0.1 127.0.0.2',
        'maximum-elapsed-time': '5',
      }
    )
    self.runAndAssertPassedMessage(
      "http://www.httpallok.com/ : "
      "dns_query: OK resolver's 1.2.3.4: 127.0.0.1 127.0.0.2 "
      "whois: OK httpallok.com expires in > 30 days "
      "tcp_server: OK IP 127.0.0.1:80 OK IP 127.0.0.2:80 "
      "http_query: OK IP 127.0.0.1 status_code 302 OK IP 127.0.0.2 "
      "status_code 302 "
      "ssl_certificate: OK No check needed "
      "elapsed_time: OK IP 127.0.0.1 replied < 5.00s OK IP 127.0.0.2 replied "
      "< 5.00s"
    )

  def test_configuration_no_ip_list(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "dns_query: OK No check configured "
      "whois: OK allok.com expires in > 30 days "
      "tcp_server: OK No check configured "
      "http_query: OK IP 127.0.0.1 status_code 302 OK IP 127.0.0.2 "
      "status_code 302 "
      "ssl_certificate: OK IP 127.0.0.1 expires in > 15 days OK IP "
      "127.0.0.2 expires in > 15 days "
      "elapsed_time: OK No check configured"
    )

  def test_all_ok_no_ssl_certificate(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
        'ip-list': '127.0.0.1 127.0.0.2',
        'maximum-elapsed-time': '5',
        'enabled-sense-list': 'dns_query whois tcp_server http_query '
                              'elapsed_time',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "dns_query: OK resolver's 1.2.3.4: 127.0.0.1 127.0.0.2 "
      "whois: OK allok.com expires in > 30 days "
      "tcp_server: OK IP 127.0.0.1:443 OK IP 127.0.0.2:443 "
      "http_query: OK IP 127.0.0.1 status_code 302 OK IP 127.0.0.2 "
      "status_code 302 "
      "elapsed_time: OK IP 127.0.0.1 replied < 5.00s OK IP 127.0.0.2 replied "
      "< 5.00s"
    )

  def test_all_ok_only_ssl_certificate(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
        'ip-list': '127.0.0.1 127.0.0.2',
        'maximum-elapsed-time': '5',
        'enabled-sense-list': 'ssl_certificate',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "ssl_certificate: OK IP 127.0.0.1 expires in > 15 days OK IP "
      "127.0.0.2 expires in > 15 days"
    )

class TestCheckSurykatkaJSONHttpQuery(CheckSurykatkaJSONMixin):
  def writeSurykatkaPromise(self, d):
    d.update(**{
        'report': 'http_query',
        'json-file': self.json_file,
    })
    super().writeSurykatkaPromise(d)

  def setUp(self):
    super().setUp()
    self.writeSurykatkaJson({
      "http_query": [
        {
          "ip": "127.0.0.1",
          "status_code": 302,
          "url": "https://www.allok.com/",
          "total_seconds": 4
        },
        {
          "ip": "127.0.0.2",
          "status_code": 302,
          "url": "https://www.allok.com/",
          "total_seconds": 4
        },
        {
          "ip": "2001:db8::8a2e:370:7334",
          "status_code": 302,
          "url": "https://www.allok.com/",
          "total_seconds": 4
        },
        {
          "ip": "fe80::b4b4:cdff:fe96:bee4",
          "status_code": 302,
          "url": "https://www.allok.com/",
          "total_seconds": 4
        },
        {
          "ip": "127.0.0.1",
          "status_code": 302,
          "url": "http://www.httpallok.com/",
          "total_seconds": 4
        },
        {
          "ip": "127.0.0.2",
          "status_code": 302,
          "url": "http://www.httpallok.com/",
          "total_seconds": 4
        },
        {
          "ip": "2001:db8::8a2e:370:7334",
          "status_code": 302,
          "url": "http://www.httpallok.com/",
          "total_seconds": 4
        },
        {
          "ip": "fe80::b4b4:cdff:fe96:bee4",
          "status_code": 302,
          "url": "http://www.httpallok.com/",
          "total_seconds": 4
        },
      ],
      "ssl_certificate": [
        {
          "hostname": "www.allok.com",
          "ip": "127.0.0.1",
          "not_after": self.time_future60d
        },
        {
          "hostname": "www.allok.com",
          "ip": "127.0.0.2",
          "not_after": self.time_future60d
        },
        {
          "hostname": "www.allok.com",
          "ip": "2001:db8::8a2e:370:7334",
          "not_after": self.time_future60d
        },
        {
          "hostname": "www.allok.com",
          "ip": "fe80::b4b4:cdff:fe96:bee4",
          "not_after": self.time_future60d
        },
      ],
      "dns_query": [
        {
            "domain": "www.allok.com",
            "rdtype": "A",
            "resolver_ip": "1.2.3.4",
            "response": "127.0.0.1, 127.0.0.2"
        },
        {
            "domain": "www.allok.com",
            "rdtype": "AAAA",
            "resolver_ip": "1.2.3.4",
            "response": "2001:db8::8a2e:370:7334, fe80::b4b4:cdff:fe96:bee4"
        },
        {
            "domain": "www.httpallok.com",
            "rdtype": "A",
            "resolver_ip": "1.2.3.4",
            "response": "127.0.0.1, 127.0.0.2"
        },
        {
            "domain": "www.httpallok.com",
            "rdtype": "AAAA",
            "resolver_ip": "1.2.3.4",
            "response": "2001:db8::8a2e:370:7334, fe80::b4b4:cdff:fe96:bee4"
        },
      ],
      "tcp_server": [
        {
            "ip": "127.0.0.1",
            "state": "open",
            "port": 443,
            "domain": "www.allok.com"
        },
        {
            "ip": "127.0.0.2",
            "state": "open",
            "port": 443,
            "domain": "www.allok.com"
        },
        {
            "ip": "2001:db8::8a2e:370:7334",
            "state": "open",
            "port": 443,
            "domain": "www.allok.com"
        },
        {
            "ip": "fe80::b4b4:cdff:fe96:bee4",
            "state": "open",
            "port": 443,
            "domain": "www.allok.com"
        },
        {
            "ip": "127.0.0.1",
            "state": "open",
            "port": 80,
            "domain": "www.httpallok.com"
        },
        {
            "ip": "127.0.0.2",
            "state": "open",
            "port": 80,
            "domain": "www.httpallok.com"
        },
        {
            "ip": "2001:db8::8a2e:370:7334",
            "state": "open",
            "port": 80,
            "domain": "www.httpallok.com"
        },
        {
            "ip": "fe80::b4b4:cdff:fe96:bee4",
            "state": "open",
            "port": 80,
            "domain": "www.httpallok.com"
        },
      ],
      "whois": [
        {
            "domain": "allok.com",
            "expiration_date": self.time_future60d,
        },
        {
            "domain": "httpallok.com",
            "expiration_date": self.time_future60d,
        },
      ]
    })

  def test_all_ok(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
        'ip-list': '127.0.0.1 127.0.0.2 2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4',
        'maximum-elapsed-time': '5',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "dns_query: OK resolver's 1.2.3.4: 127.0.0.1 127.0.0.2 "
        "2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4 "
      "whois: OK allok.com expires in > 30 days "
      "tcp_server: OK IP 127.0.0.1:443 OK IP 127.0.0.2:443 "
        "OK IP [2001:db8::8a2e:370:7334]:443 OK IP [fe80::b4b4:cdff:fe96:bee4]:443 "
      "http_query: OK IP 127.0.0.1 status_code 302 OK IP 127.0.0.2 "
        "status_code 302 OK IP 2001:db8::8a2e:370:7334 status_code 302 "
        "OK IP fe80::b4b4:cdff:fe96:bee4 status_code 302 "
      "ssl_certificate: OK IP 127.0.0.1 expires in > 15 days OK IP "
        "127.0.0.2 expires in > 15 days OK IP 2001:db8::8a2e:370:7334"
        " expires in > 15 days OK IP fe80::b4b4:cdff:fe96:bee4 expires in > 15 days "
      "elapsed_time: OK IP 127.0.0.1 replied < 5.00s OK IP 127.0.0.2 replied "
        "< 5.00s OK IP 2001:db8::8a2e:370:7334 replied < 5.00s "
        "OK IP fe80::b4b4:cdff:fe96:bee4 replied < 5.00s"
    )

  def test_http_all_ok(self):
    self.writeSurykatkaPromise(
      {
        'report': 'http_query',
        'json-file': self.json_file,
        'url': 'http://www.httpallok.com/',
        'status-code': '302',
        'ip-list': '127.0.0.1 127.0.0.2 2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4',
        'maximum-elapsed-time': '5',
      }
    )
    self.runAndAssertPassedMessage(
      "http://www.httpallok.com/ : "
      "dns_query: OK resolver's 1.2.3.4: 127.0.0.1 127.0.0.2 "
        "2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4 "
      "whois: OK httpallok.com expires in > 30 days "
      "tcp_server: OK IP 127.0.0.1:80 OK IP 127.0.0.2:80 "
        "OK IP [2001:db8::8a2e:370:7334]:80 OK IP [fe80::b4b4:cdff:fe96:bee4]:80 "
      "http_query: OK IP 127.0.0.1 status_code 302 OK IP 127.0.0.2 "
        "status_code 302 OK IP 2001:db8::8a2e:370:7334 status_code 302 "
        "OK IP fe80::b4b4:cdff:fe96:bee4 status_code 302 "
      "ssl_certificate: OK No check needed "
      "elapsed_time: OK IP 127.0.0.1 replied < 5.00s OK IP 127.0.0.2 replied "
        "< 5.00s OK IP 2001:db8::8a2e:370:7334 replied < 5.00s "
        "OK IP fe80::b4b4:cdff:fe96:bee4 replied < 5.00s"
    )

  def test_configuration_no_ip_list(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "dns_query: OK No check configured "
      "whois: OK allok.com expires in > 30 days "
      "tcp_server: OK No check configured "
      "http_query: OK IP 127.0.0.1 status_code 302 OK IP 127.0.0.2 "
        "status_code 302 OK IP 2001:db8::8a2e:370:7334 status_code 302 "
        "OK IP fe80::b4b4:cdff:fe96:bee4 status_code 302 "
      "ssl_certificate: OK IP 127.0.0.1 expires in > 15 days OK IP "
        "127.0.0.2 expires in > 15 days OK IP 2001:db8::8a2e:370:7334"
        " expires in > 15 days OK IP fe80::b4b4:cdff:fe96:bee4 expires in > 15 days "
      "elapsed_time: OK No check configured"
    )

  def test_all_ok_nothing_enabled(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
        'ip-list': '127.0.0.1 127.0.0.2',
        'maximum-elapsed-time': '5',
        'enabled-sense-list': '',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ :"
    )

  def test_all_ok_no_ssl_certificate(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
        'ip-list': '127.0.0.1 127.0.0.2 2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4',
        'maximum-elapsed-time': '5',
        'enabled-sense-list': 'dns_query whois tcp_server http_query '
                              'elapsed_time',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "dns_query: OK resolver's 1.2.3.4: 127.0.0.1 127.0.0.2 "
        "2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4 "
      "whois: OK allok.com expires in > 30 days "
      "tcp_server: OK IP 127.0.0.1:443 OK IP 127.0.0.2:443 "
        "OK IP [2001:db8::8a2e:370:7334]:443 OK IP [fe80::b4b4:cdff:fe96:bee4]:443 "
      "http_query: OK IP 127.0.0.1 status_code 302 OK IP 127.0.0.2 "
        "status_code 302 OK IP 2001:db8::8a2e:370:7334 status_code 302 "
        "OK IP fe80::b4b4:cdff:fe96:bee4 status_code 302 "
      "elapsed_time: OK IP 127.0.0.1 replied < 5.00s OK IP 127.0.0.2 replied "
        "< 5.00s OK IP 2001:db8::8a2e:370:7334 replied < 5.00s "
        "OK IP fe80::b4b4:cdff:fe96:bee4 replied < 5.00s"
    )

  def test_all_ok_only_ssl_certificate(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
        'ip-list': '127.0.0.1 127.0.0.2',
        'maximum-elapsed-time': '5',
        'enabled-sense-list': 'ssl_certificate',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "ssl_certificate: OK IP 127.0.0.1 expires in > 15 days OK IP "
        "127.0.0.2 expires in > 15 days OK IP 2001:db8::8a2e:370:7334"
        " expires in > 15 days OK IP fe80::b4b4:cdff:fe96:bee4 expires in > 15 days"
    )

class TestCheckSurykatkaJSONHttpQueryOnlyIPV6(CheckSurykatkaJSONMixin):
  def writeSurykatkaPromise(self, d):
    d.update(**{
        'report': 'http_query',
        'json-file': self.json_file,
    })
    super().writeSurykatkaPromise(d)

  def setUp(self):
    super().setUp()
    self.writeSurykatkaJson({
      "http_query": [
        {
          "ip": "2001:db8::8a2e:370:7334",
          "status_code": 302,
          "url": "https://www.allok.com/",
          "total_seconds": 4
        },
        {
          "ip": "fe80::b4b4:cdff:fe96:bee4",
          "status_code": 302,
          "url": "https://www.allok.com/",
          "total_seconds": 4
        },
        {
          "ip": "2001:db8::8a2e:370:7334",
          "status_code": 302,
          "url": "http://www.httpallok.com/",
          "total_seconds": 4
        },
        {
          "ip": "fe80::b4b4:cdff:fe96:bee4",
          "status_code": 302,
          "url": "http://www.httpallok.com/",
          "total_seconds": 4
        },
      ],
      "ssl_certificate": [
        {
          "hostname": "www.allok.com",
          "ip": "2001:db8::8a2e:370:7334",
          "not_after": self.time_future60d
        },
        {
          "hostname": "www.allok.com",
          "ip": "fe80::b4b4:cdff:fe96:bee4",
          "not_after": self.time_future60d
        },
      ],
      "dns_query": [
        {
            "domain": "www.allok.com",
            "rdtype": "AAAA",
            "resolver_ip": "1.2.3.4",
            "response": "2001:db8::8a2e:370:7334, fe80::b4b4:cdff:fe96:bee4"
        },
        {
            "domain": "www.httpallok.com",
            "rdtype": "AAAA",
            "resolver_ip": "1.2.3.4",
            "response": "2001:db8::8a2e:370:7334, fe80::b4b4:cdff:fe96:bee4"
        },
      ],
      "tcp_server": [
        {
            "ip": "2001:db8::8a2e:370:7334",
            "state": "open",
            "port": 443,
            "domain": "www.allok.com"
        },
        {
            "ip": "fe80::b4b4:cdff:fe96:bee4",
            "state": "open",
            "port": 443,
            "domain": "www.allok.com"
        },
        {
            "ip": "2001:db8::8a2e:370:7334",
            "state": "open",
            "port": 80,
            "domain": "www.httpallok.com"
        },
        {
            "ip": "fe80::b4b4:cdff:fe96:bee4",
            "state": "open",
            "port": 80,
            "domain": "www.httpallok.com"
        },
      ],
      "whois": [
        {
            "domain": "allok.com",
            "expiration_date": self.time_future60d,
        },
        {
            "domain": "httpallok.com",
            "expiration_date": self.time_future60d,
        },
      ]
    })

  def test_all_ok(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
        'ip-list': '2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4',
        'maximum-elapsed-time': '5',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "dns_query: OK resolver's 1.2.3.4: 2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4 "
      "whois: OK allok.com expires in > 30 days "
      "tcp_server: OK IP [2001:db8::8a2e:370:7334]:443 OK IP [fe80::b4b4:cdff:fe96:bee4]:443 "
      "http_query: OK IP 2001:db8::8a2e:370:7334 status_code 302 "
        "OK IP fe80::b4b4:cdff:fe96:bee4 status_code 302 "
      "ssl_certificate: OK IP 2001:db8::8a2e:370:7334"
        " expires in > 15 days OK IP fe80::b4b4:cdff:fe96:bee4 expires in > 15 days "
      "elapsed_time: OK IP 2001:db8::8a2e:370:7334 replied < 5.00s "
        "OK IP fe80::b4b4:cdff:fe96:bee4 replied < 5.00s"
    )

  def test_http_all_ok(self):
    self.writeSurykatkaPromise(
      {
        'report': 'http_query',
        'json-file': self.json_file,
        'url': 'http://www.httpallok.com/',
        'status-code': '302',
        'ip-list': '2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4',
        'maximum-elapsed-time': '5',
      }
    )
    self.runAndAssertPassedMessage(
      "http://www.httpallok.com/ : "
      "dns_query: OK resolver's 1.2.3.4: 2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4 "
      "whois: OK httpallok.com expires in > 30 days "
      "tcp_server: OK IP [2001:db8::8a2e:370:7334]:80 OK IP [fe80::b4b4:cdff:fe96:bee4]:80 "
      "http_query: OK IP 2001:db8::8a2e:370:7334 status_code 302 "
        "OK IP fe80::b4b4:cdff:fe96:bee4 status_code 302 "
      "ssl_certificate: OK No check needed "
      "elapsed_time: OK IP 2001:db8::8a2e:370:7334 replied < 5.00s "
        "OK IP fe80::b4b4:cdff:fe96:bee4 replied < 5.00s"
    )

  def test_configuration_no_ip_list(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "dns_query: OK No check configured "
      "whois: OK allok.com expires in > 30 days "
      "tcp_server: OK No check configured "
      "http_query: OK IP 2001:db8::8a2e:370:7334 status_code 302 "
        "OK IP fe80::b4b4:cdff:fe96:bee4 status_code 302 "
      "ssl_certificate: OK IP 2001:db8::8a2e:370:7334"
        " expires in > 15 days OK IP fe80::b4b4:cdff:fe96:bee4 expires in > 15 days "
      "elapsed_time: OK No check configured"
    )


  def test_all_ok_no_ssl_certificate(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
        'ip-list': '2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4',
        'maximum-elapsed-time': '5',
        'enabled-sense-list': 'dns_query whois tcp_server http_query '
                              'elapsed_time',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "dns_query: OK resolver's 1.2.3.4: 2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4 "
      "whois: OK allok.com expires in > 30 days "
      "tcp_server: OK IP [2001:db8::8a2e:370:7334]:443 OK IP [fe80::b4b4:cdff:fe96:bee4]:443 "
      "http_query: OK IP 2001:db8::8a2e:370:7334 status_code 302 "
        "OK IP fe80::b4b4:cdff:fe96:bee4 status_code 302 "
      "elapsed_time: OK IP 2001:db8::8a2e:370:7334 replied < 5.00s "
        "OK IP fe80::b4b4:cdff:fe96:bee4 replied < 5.00s"
    )

  def test_all_ok_only_ssl_certificate(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.allok.com/',
        'status-code': '302',
        'ip-list': '2001:db8::8a2e:370:7334 fe80::b4b4:cdff:fe96:bee4',
        'maximum-elapsed-time': '5',
        'enabled-sense-list': 'ssl_certificate',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.allok.com/ : "
      "ssl_certificate: OK IP 2001:db8::8a2e:370:7334"
        " expires in > 15 days OK IP fe80::b4b4:cdff:fe96:bee4 expires in > 15 days"
    )


class TestCheckSurykatkaJSONHttpQueryDnsQuery(CheckSurykatkaJSONMixin):
  def writeSurykatkaPromise(self, d):
    d.update(**{
        'report': 'http_query',
        'json-file': self.json_file,
        'enabled-sense-list': 'dns_query',
    })
    super().writeSurykatkaPromise(d)

  def setUp(self):
    super().setUp()
    self.writeSurykatkaJson({
      "dns_query": [
        {
            "domain": "www.httpallok.com",
            "rdtype": "A",
            "resolver_ip": "1.2.3.4",
            "response": "127.0.0.1, 127.0.0.2"
        },
        {
            "domain": "www.badip.com",
            "rdtype": "A",
            "resolver_ip": "1.2.3.4",
            "response": "127.0.0.1, 127.0.0.4"
        },
        {
            "domain": "www.dnsquerynoreply.com",
            "rdtype": "A",
            "resolver_ip": "1.2.3.4",
            "response": ""
        },
      ],
    })

  def test_bad_ip(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.badip.com/',
        'ip-list': '127.0.0.1 127.0.0.2',
      }
    )
    self.configureLauncher(enable_anomaly=True)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "http://www.badip.com/ : "
      "dns_query: ERROR resolver's 1.2.3.4: 127.0.0.1 127.0.0.2 != "
      "127.0.0.1 127.0.0.4"
    )

  def test_no_entry(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.dnsquerynoentry.com/',
        'ip-list': '127.0.0.1',
      }
    )
    self.runAndAssertFailedMessage(
      "http://www.dnsquerynoentry.com/ : "
      "dns_query: ERROR No data"
    )

  def test_query_no_key(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.dnsquerynokey.com/',
        'ip-list': '127.0.0.1',
      }
    )
    self.writeSurykatkaJson({})
    self.runAndAssertFailedMessage(
      "http://www.dnsquerynokey.com/ : "
      "dns_query: ERROR 'dns_query' not in %(json_file)r" % {
        'json_file': self.json_file}
    )

  def test_mismatch(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.httpallok.com/',
        'ip-list': '127.0.0.1 127.0.0.9',
      }
    )
    self.runAndAssertFailedMessage(
      "http://www.httpallok.com/ : "
      "dns_query: ERROR resolver's 1.2.3.4: 127.0.0.1 127.0.0.9 != "
      "127.0.0.1 127.0.0.2"
    )

  def test_no_reply(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.dnsquerynoreply.com/',
        'ip-list': '127.0.0.1',
      }
    )
    self.runAndAssertFailedMessage(
      "http://www.dnsquerynoreply.com/ : "
      "dns_query: ERROR resolver's 1.2.3.4: 127.0.0.1 != empty-reply"
    )


class TestCheckSurykatkaJSONHttpQueryWhois(CheckSurykatkaJSONMixin):
  def writeSurykatkaPromise(self, d):
    d.update(**{
        'report': 'http_query',
        'json-file': self.json_file,
        'enabled-sense-list': 'whois',
    })
    super().writeSurykatkaPromise(d)

  def setUp(self):
    super().setUp()
    self.writeSurykatkaJson({
      "whois": [
        {
            "domain": "whois3.com",
            "expiration_date": self.time_future3d,
        },
        {
            "domain": "whois29.com",
            "expiration_date": self.time_future29d
        },
        {
            "domain": "whoisminus29.com",
            "expiration_date": self.time_past29d
        },
        {
            "domain": "whoisnone.com",
            "expiration_date": None
        },
      ]
    })

  def test_no_entry(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.whoisnoentry.com/',
        'enabled-sense-list': 'whois',
      }
    )
    self.runAndAssertFailedMessage(
      "http://www.whoisnoentry.com/ : "
      "whois: ERROR No data"
    )

  def test_no_key(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.whoisnokey.com/',
      }
    )
    self.writeSurykatkaJson({
      "dns_query": [
      ],
    })
    self.runAndAssertFailedMessage(
      "http://www.whoisnokey.com/ : "
      "whois: ERROR 'whois' not in %(json_file)r" % {
        'json_file': self.json_file}
    )

  def test_expires_2_day(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.whois3.com/',
        'domain-expiration-days': '2',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.whois3.com/ : "
      "whois: OK whois3.com expires in > 2 days"
    )

  def test_expired_expires_2_day(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.whois3.com/',
        'domain-expiration-days': '4',
      }
    )
    self.runAndAssertFailedMessage(
      "https://www.whois3.com/ : "
      "whois: ERROR whois3.com expires in < 4 days"
    )

  def test_expired(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.whois29.com/',
      }
    )
    self.runAndAssertFailedMessage(
      "https://www.whois29.com/ : "
      "whois: ERROR whois29.com expires in < 30 days"
    )

  def test_expired_before_today(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.whoisminus29.com/',
      }
    )
    self.runAndAssertFailedMessage(
      "https://www.whoisminus29.com/ : "
      "whois: ERROR whoisminus29.com expires in < 30 days"
    )

  def test_none(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.whoisnone.com/',
      }
    )
    self.runAndAssertFailedMessage(
      "https://www.whoisnone.com/ : "
      "whois: ERROR whoisnone.com expiration date not avaliable"
    )


class TestCheckSurykatkaJSONHttpQueryTcpServer(CheckSurykatkaJSONMixin):
  def setUp(self):
    super().setUp()
    self.writeSurykatkaJson({
      "tcp_server": [
        {
            "ip": "127.0.0.2",
            "state": "open",
            "port": 80,
            "domain": "www.tcpservernoip.com"
        },
        {
            "ip": "127.0.0.1",
            "state": "filtered",
            "port": 80,
            "domain": "www.tcpserverfiltered.com"
        },
      ]
    })

  def writeSurykatkaPromise(self, d):
    d.update(**{
        'report': 'http_query',
        'json-file': self.json_file,
        'enabled-sense-list': 'tcp_server',
    })
    super().writeSurykatkaPromise(d)

  def test_tcp_server_no_ip(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.tcpservernoip.com/',
        'ip-list': '127.0.0.1',
      }
    )
    self.runAndAssertFailedMessage(
      "http://www.tcpservernoip.com/ : "
      "tcp_server: ERROR IP 127.0.0.1:80"
    )

  def test_tcp_server_filtered(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.tcpserverfiltered.com/',
        'ip-list': '127.0.0.1',
      }
    )
    self.runAndAssertFailedMessage(
      "http://www.tcpserverfiltered.com/ : "
      "tcp_server: ERROR IP 127.0.0.1:80"
    )

  def test_tcp_server_no_entry(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.tcpservernoentry.com/',
        'ip-list': '127.0.0.1',
      }
    )
    self.runAndAssertFailedMessage(
      "http://www.tcpservernoentry.com/ : "
      "tcp_server: ERROR No data"
    )

  def test_tcp_server_no_key(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.tcpservernokey.com/',
        'ip-list': '127.0.0.1',
      }
    )
    self.writeSurykatkaJson({
      "dns_query": [
      ],
    })
    self.runAndAssertFailedMessage(
      "http://www.tcpservernokey.com/ : "
      "tcp_server: ERROR 'tcp_server' not in %(json_file)r" % {
        'json_file': self.json_file}
    )


class TestCheckSurykatkaJSONHttpQueryHttpQuery(CheckSurykatkaJSONMixin):
  def setUp(self):
    super().setUp()
    self.writeSurykatkaJson({
      "http_query": [
        {
          "ip": "127.0.0.1",
          "status_code": 302,
          "url": "http://www.httpallok.com/",
          "total_seconds": 4
        },
        {
          "ip": "127.0.0.2",
          "status_code": 302,
          "url": "http://www.httpallok.com/",
          "total_seconds": 4
        },
        {
          "ip": "127.0.0.1",
          "status_code": 200,
          "url": "http://www.httpheader.com/",
          "http_header_dict": {
            "Vary": "Accept-Encoding", "Cache-Control": "max-age=300, public"},
        },
        {
          "ip": "127.0.0.1",
          "status_code": 302,
          "url": "http://www.badip.com/",
        },
        {
          "ip": "127.0.0.4",
          "status_code": 302,
          "url": "http://www.badip.com/",
        },
      ],
    })

  def writeSurykatkaPromise(self, d):
    d.update(**{
        'report': 'http_query',
        'json-file': self.json_file,
        'enabled-sense-list': 'http_query',
    })
    super().writeSurykatkaPromise(d)

  def _test_bad_code_explanation(self, status_code, explanation):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.statuscode.com/',
        'status-code': '301',
      }
    )
    self.writeSurykatkaJson({
      "http_query": [
        {
          "ip": "127.0.0.1",
          "status_code": status_code,
          "url": "http://www.statuscode.com/"
        }
      ],
    })
    self.runAndAssertFailedMessage(
      "http://www.statuscode.com/ : "
      "http_query: ERROR IP 127.0.0.1 status_code %s != 301" % (explanation,)
    )

  def test_bad_code_explanation_520(self):
    self._test_bad_code_explanation(520, '520 (Too many redirects)')

  def test_bad_code_explanation_523(self):
    self._test_bad_code_explanation(523, '523 (Connection error)')

  def test_bad_code_explanation_524(self):
    self._test_bad_code_explanation(524, '524 (Connection timeout)')

  def test_bad_code_explanation_526(self):
    self._test_bad_code_explanation(526, '526 (SSL Error)')

  def test_bad_code(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.httpallok.com/',
        'status-code': '301',
      }
    )
    self.runAndAssertFailedMessage(
      "http://www.httpallok.com/ : "
      "http_query: ERROR IP 127.0.0.1 status_code 302 != 301 ERROR "
      "IP 127.0.0.2 status_code 302 != 301"
    )

  def test_not_present(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.httpquerynopresent.com/',
        'status-code': '302',
      }
    )
    self.writeSurykatkaJson({
    })
    self.runAndAssertFailedMessage(
      "http://www.httpquerynopresent.com/ : "
      "http_query: ERROR 'http_query' not in %(json_file)r" % {
        'json_file': self.json_file}
    )

  def test_no_data(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.httpquerynodata.com/',
        'status-code': '302',
      }
    )
    self.runAndAssertFailedMessage(
      "http://www.httpquerynodata.com/ : "
      "http_query: ERROR No data"
    )

  def test_header_dict_mismatch(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.httpheader.com/',
        'status-code': '200',
        'http-header-dict': '{"Vary": "Accept-Encoding", "Cache-Control": '
        '"max-age=300"}',
      }
    )
    self.runAndAssertFailedMessage(
      'http://www.httpheader.com/ : '
      'http_query: OK IP 127.0.0.1 status_code 200 ERROR IP 127.0.0.1 '
      'HTTP Header {"Cache-Control": "max-age=300", "Vary": '
      '"Accept-Encoding"} != {"Cache-Control": "max-age=300, public", "Vary": '
      '"Accept-Encoding"}'
    )

  def test_header_dict(self):
    self.writeSurykatkaPromise(
      {
        'url': 'http://www.httpheader.com/',
        'status-code': '200',
        'http-header-dict': '{"Vary": "Accept-Encoding", "Cache-Control": '
        '"max-age=300, public"}',
      }
    )
    self.runAndAssertPassedMessage(
      'http://www.httpheader.com/ : '
      'http_query: OK IP 127.0.0.1 status_code 200 OK IP 127.0.0.1 HTTP '
      'Header {"Cache-Control": "max-age=300, public", "Vary": '
      '"Accept-Encoding"}'
    )


class TestCheckSurykatkaJSONHttpQuerySslCertificate(CheckSurykatkaJSONMixin):
  def setUp(self):
    super().setUp()
    self.writeSurykatkaJson({
      "ssl_certificate": [
        {
          "hostname": "www.cert3.com",
          "ip": "127.0.0.1",
          "not_after": self.time_future3d
        },
        {
          "hostname": "www.cert14.com",
          "ip": "127.0.0.1",
          "not_after": self.time_future14d
        },
        {
          "hostname": "www.certminus14.com",
          "ip": "127.0.0.1",
          "not_after": self.time_past14d
        },
        {
          "hostname": "www.sslcertnoinfo.com",
          "ip": "127.0.0.1",
          "not_after": None
        },
      ],
    })

  def writeSurykatkaPromise(self, d):
    d.update(**{
        'report': 'http_query',
        'json-file': self.json_file,
        'enabled-sense-list': 'ssl_certificate',
    })
    super().writeSurykatkaPromise(d)

  def test_good_certificate_2_day(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.cert3.com/',
        'certificate-expiration-days': '2',
      }
    )
    self.runAndAssertPassedMessage(
      "https://www.cert3.com/ : "
      "ssl_certificate: OK IP 127.0.0.1 expires in > 2 days"
    )

  def test_expired_certificate_4_day(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.cert3.com/',
        'certificate-expiration-days': '4',
      }
    )

    self.runAndAssertFailedMessage(
      "https://www.cert3.com/ : "
      "ssl_certificate: ERROR IP 127.0.0.1 expires in < 4 days"
    )

  def test_expired_certificate(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.cert14.com/',
      }
    )
    self.runAndAssertFailedMessage(
      "https://www.cert14.com/ : "
      "ssl_certificate: ERROR IP 127.0.0.1 expires in < 15 days"
    )

  def test_expired_certificate_before_today(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.certminus14.com/',
      }
    )
    self.runAndAssertFailedMessage(
      "https://www.certminus14.com/ : "
      "ssl_certificate: ERROR IP 127.0.0.1 expires in < 15 days"
    )

  def test_https_no_cert(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.sslcertnoinfo.com/',
      }
    )
    self.runAndAssertFailedMessage(
      "https://www.sslcertnoinfo.com/ : "
      "ssl_certificate: ERROR IP 127.0.0.1 no information"
    )

  def test_no_ssl_certificate(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.nosslcertificate.com/',
      }
    )
    self.writeSurykatkaJson({
      "http_query": [
        {
          "ip": "127.0.0.1",
          "status_code": 302,
          "url": "https://www.nosslcertificate.com/"
        },
      ],
    })
    self.runAndAssertFailedMessage(
      "https://www.nosslcertificate.com/ : "
      "ssl_certificate: ERROR 'ssl_certificate' not in %(json_file)r" % {
        'json_file': self.json_file}
    )


class TestCheckSurykatkaJSONHttpQueryElapsedTime(CheckSurykatkaJSONMixin):
  def writeSurykatkaPromise(self, d):
    d.update(**{
        'report': 'http_query',
        'json-file': self.json_file,
        'enabled-sense-list': 'elapsed_time',
    })
    super().writeSurykatkaPromise(d)

  def setUp(self):
    super().setUp()
    self.writeSurykatkaJson({
      "http_query": [
        {
          "ip": "127.0.0.1",
          "status_code": 302,
          "url": "https://www.elapsedtoolong.com/",
          "total_seconds": 6
        },
        {
          "ip": "127.0.0.1",
          "status_code": 302,
          "url": "https://www.elapsednototal.com/",
        },
      ]
    })

  def test_too_long(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.elapsedtoolong.com/',
        'ip-list': '127.0.0.1',
        'maximum-elapsed-time': '5',
      }
    )
    self.runAndAssertFailedMessage(
      "https://www.elapsedtoolong.com/ : "
      "elapsed_time: ERROR IP 127.0.0.1 replied > 5.00s"
    )

  def test_no_match(self):
    self.writeSurykatkaPromise(
      {
        'url': 'https://www.elapsednototal.com/',
        'ip-list': '127.0.0.1',
        'maximum-elapsed-time': '5',
      }
    )
    self.runAndAssertFailedMessage(
      "https://www.elapsednototal.com/ : "
      "elapsed_time: ERROR No entry with total_seconds found. If the error "
      "persist, please update surykatka"
    )



from __future__ import division

from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise

import os
import sys

import sqlite3
import argparse
import datetime
import psutil
import itertools
import warnings
import pkgutil

from slapos.collect.db import Database
from contextlib import closing

# install pandas, numpy and statsmodels for ARIMA prediction
try:
  import pandas as pd
  import numpy as np
  from statsmodels.tsa.arima.model import ARIMA
except ImportError:
  pass

@implementer(interface.IPromise)
class RunPromise(GenericPromise):

  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # check disk space at least every hours (heavy in computation)
    self.setPeriodicity(float(self.getConfig('frequency', 60)))

  def getDiskSize(self, disk_partition, database):
    database = Database(database, create=False, timeout=10)
    # by using contextlib.closing, we don't need to close the database explicitly
    with closing(database):
      try:
        # fetch disk size
        database.connect()
        where_query = "partition='%s'" % (disk_partition)
        order = "datetime(date || ' ' || time) DESC"
        query_result = database.select("disk", columns="free+used", where=where_query, order=order, limit=1)
        result = query_result.fetchone()
        if not result or not result[0]:
          return None
        disk_size = result[0]
      except sqlite3.OperationalError as e:
        # if database is still locked after timeout expiration (another process is using it)
        # we print warning message and try the promise at next run until max warn count
        locked_message = "database is locked"
        if locked_message in str(e) and \
            not self.raiseOnDatabaseLocked(locked_message):
          return None
        raise
    return disk_size

  def getFreeSpace(self, disk_partition, database, date, time):
    database = Database(database, create=False, timeout=10)
    with closing(database):
      try:
        # fetch free disk space
        database.connect()
        where_query = "time between '%s:00' and '%s:30' and partition='%s'" % (time, time, disk_partition)
        query_result = database.select("disk", date, "free", where=where_query)
        result = query_result.fetchone()
        if not result or not result[0]:
          self.logger.info("No result from collector database: disk check skipped")
          return -1
        disk_free = result[0]
      except sqlite3.OperationalError as e:
        # if database is still locked after timeout expiration (another process is using it)
        # we print warning message and try the promise at next run until max warn count
        locked_message = "database is locked"
        if locked_message in str(e) and \
            not self.raiseOnDatabaseLocked(locked_message):
          return -1
        raise
    return int(disk_free)

  def getBiggestPartitions(self, database, date, time):
    # displays the 3 biggest partitions thanks to disk usage
    limit = 3
    database = Database(database, create=False, timeout=10)
    with closing(database):
      try:
        database.connect()
        date_time = date + ' ' + time
        # gets the data recorded between the current date (date_time) and 24 hours earlier
        where_query = "datetime(date || ' ' || time) >= datetime('%s', '-1 days') AND datetime(date || ' ' || time) <= datetime('%s')"
        # gets only the most recent data for each partition
        result = database.select(
          "folder",
          columns = "partition, disk_used*1024, max(datetime(date || ' ' || time))",
          where =  where_query % (date_time, date_time),
          group = "partition",
          order = "disk_used DESC",
          limit = limit).fetchall()
        if not result or not result[0]:
          self.logger.info("No result from collector database in table folder: skipped")
          return None
      except sqlite3.OperationalError as e:
        # if database is still locked after timeout expiration (another process is using it)
        # we print warning message and try the promise at next run until max warn count
        locked_message = "database is locked"
        if locked_message in str(e) and \
            not self.raiseOnDatabaseLocked(locked_message):
          return None
        raise
    return result

  def diskSpacePrediction(self, disk_partition, database, date, time, day_range):
    """
    Returns an estimation of free disk space left depending on
    the day_range parameter.

    It uses Arima in order to predict data thanks to the 15 days before.
    """
    database = Database(database, create=False, timeout=10)
    with closing(database):
      try:
        database.connect()
        # get one data per day, where each data is at the same time
        where_query = "time between '%s:00' and '%s:30' and partition='%s'" % (
          time, time, disk_partition)
        result = database.select(
          "disk",
          columns = "free, datetime(date || ' ' || time)",
          where = where_query,
          order = "datetime(date || ' ' || time) ASC").fetchall()
        # checks that there are at least 14 days of data
        if (not result) or (len(result) < 14):
          self.logger.info("No or not enough results from collector database in table disk: no prediction")
          return None
        # put the list in pandas dataframe format and set the right types
        data = np.array(result,
                dtype=[('free', 'float'), ('date', 'datetime64[s]')])
        df = pd.DataFrame.from_records(data)
        df.loc[:,'date'] = pd.to_datetime(df.date)
        df = df.set_index('date')
        df.index = pd.DatetimeIndex(df.index).to_period('D')
        # find the best configuration by trying different combinations
        p_values = d_values = q_values = range(0, 3)
        # We were using a function called evaluateModels() to select the
        # best ARIMA (p, d, q) order, but it evaluates 27 combinations,
        # each taking about 1–2 seconds, which exceeds the 20s limit for a promise.
        # And also in statsmodels 0.11.1, _check_estimable would skip combinations
        # with insufficient degrees of freedom.
        # but in 0.14.4, all 27 combinations are evaluated, resulting in excessive runtime.
        # To save time, we simply use the order (1, 1, 0).
        # XXX: Now we are using the hardcoded order, need to find a better way in future.
        best_cfg = (1, 1, 0)
        # set the days to be predicted
        max_date_predicted = day_range+1
        try:
          # disabling warnings during the ARIMA calculation
          model_arima = ARIMA(df, order=best_cfg, trend="t")
          model_arima_fit = model_arima.fit()
          # save ARIMA predictions
          fcast_result  = model_arima_fit.get_forecast(steps=max_date_predicted)
          fcast = fcast_result.predicted_mean
          conf = fcast_result.conf_int(alpha=0.05).to_numpy()
          if fcast.empty:
            self.logger.info("Arima prediction: none. Skipped prediction")
            return None
        except Exception:
          self.logger.info("Arima prediction error: skipped prediction")
          return None
        # get results with 95% confidence
        lower_series = conf[:, 0]
        upper_series = conf[:, 1]
        return fcast, lower_series, upper_series
      except sqlite3.OperationalError as e:
        # if database is still locked after timeout expiration (another process is using it)
        # we print warning message and try the promise at next run until max warn count
        locked_message = "database is locked"
        if locked_message in str(e) and \
            not self.raiseOnDatabaseLocked(locked_message):
          return None
        raise

  def raiseOnDatabaseLocked(self, locked_message):
    max_warn = 10
    latest_result_list = self.getLastPromiseResultList(result_count=max_warn)
    warning_count = 0
    if len(latest_result_list) < max_warn:
      return False

    for result in latest_result_list[0]:
      if result['status'] == "ERROR" and locked_message in result["message"]:
        return True

    for result_list in latest_result_list:
      found = False
      for result in result_list:
        if result['status'] == "WARNING" and locked_message in result["message"]:
          found = True
          warning_count += 1
          break
      if not found:
        break
    if warning_count == max_warn:
      # too many warning on database locked, now fail.
      return True

    self.logger.warn("collector database is locked by another process")
    return False

  @staticmethod
  def _checkInodeUsage(path):
    stat = os.statvfs(path)
    total_inode = stat.f_files
    if total_inode:
      usage = 100 * (total_inode - stat.f_ffree) / total_inode
      if usage >= 98:
        return "Disk Inodes usage is really high: %.4f%%" % usage

  def getInodeUsage(self, path):
    return (self._checkInodeUsage(path) or
       os.path.ismount('/tmp') and self._checkInodeUsage('/tmp') or
       "")

  def sense(self):
    # find if a disk is mounted on the path
    disk_partition = ""
    db_path = self.getConfig('collectordb')
    check_date = self.getConfig('test-check-date')
    path = os.path.join(self.getPartitionFolder(), "") + "extrafolder"
    partitions = psutil.disk_partitions()
    while path != '/':
      if not disk_partition:
        path = os.path.dirname(path)
      else:
        break
      for p in partitions:
        if p.mountpoint == path:
          disk_partition = p.device
          break
    if not disk_partition:
      self.logger.error("Couldn't find disk partition")
      return

    if db_path.endswith("collector.db"):
      db_path=db_path[:-len("collector.db")]

    if check_date:
      # testing mode
      currentdate = check_date
      currenttime = self.getConfig('test-check-time', '09:17')
      disk_partition = self.getConfig('test-disk-partition', '/dev/sda1')
    else:
      # get last minute
      now = datetime.datetime.utcnow()
      currentdate = now.strftime('%Y-%m-%d')
      currenttime = now - datetime.timedelta(minutes=1)
      currenttime = currenttime.time().strftime('%H:%M')

    disk_size = self.getDiskSize(disk_partition, db_path)
    # threshold is in GB
    default_threshold = 100.0
    if disk_size is not None:
      # if we know the disk size, default threshold is 5%
      default_threshold = round(disk_size/(1024*1024*1024) * 0.05, 2)
    threshold = float(self.getConfig('threshold', default_threshold))

    display_partition = bool(self.getConfig('display-partition', 0))
    if display_partition:
      # Always display the 3 partitions that have the most storage capacity on the disk
      big_partitions = self.getBiggestPartitions(db_path, currentdate, currenttime)
      if big_partitions is not None:
        for partition in big_partitions:
          user_name, size_partition, date_checked = partition
          partition_id = self.getConfig('partition-id', 'slappart')
          # get the name of each partition by adding the user's number to the general name of the partition
          partition_name = ''.join(x for x in partition_id if not x.isdigit()) + ''.join(filter(str.isdigit, user_name))
          self.logger.info("The partition %s uses %.2f G (date checked: %s).",
            partition_name, size_partition/(1024*1024*1024), date_checked)

    inode_usage = self.getInodeUsage(self.getPartitionFolder())
    if inode_usage:
      self.logger.error(inode_usage)

    free_space = self.getFreeSpace(disk_partition, db_path, currentdate,
                                   currenttime)
    if free_space == -1:
      # we couldn't connect to the database, simply ignore this occurence of sense
      return
    elif free_space > threshold*1024*1024*1024:
      self.logger.info("Current disk usage: OK")
      # if the option is enabled and the current disk size is OK,
      # we check the predicted remaining disk space
      display_prediction = bool(int(self.getConfig('display-prediction', 0) or 0))
      if display_prediction:
        # check that the libraries are installed from the slapos.toolbox extra requires
        pandas_found = pkgutil.find_loader("pandas")
        numpy_found = pkgutil.find_loader("numpy")
        statsmodels_found = pkgutil.find_loader("statsmodels")
        if pandas_found is None or numpy_found is None or statsmodels_found is None:
          self.logger.warning("Trying to use statsmodels and pandas " \
            "but at least one module is not installed. Prediction skipped.")
          return
        nb_days_predicted = int(self.getConfig('nb-days-predicted', 10) or 10)
        disk_space_prediction_tuple = self.diskSpacePrediction(
          disk_partition, db_path, currentdate, currenttime, nb_days_predicted)
        if disk_space_prediction_tuple is not None:
          fcast, lower_series, upper_series = disk_space_prediction_tuple
          space_left_predicted = fcast.iloc[-1]
          last_date_predicted = datetime.datetime.strptime(str(fcast.index[-1]),
                                                          "%Y-%m-%d")
          delta_days = (last_date_predicted.date() - \
            datetime.datetime.strptime(currentdate, "%Y-%m-%d").date()).days
          self.logger.info("Prediction: there will be %.2f G left on %s (%s days).",
            space_left_predicted/(1024*1024*1024), last_date_predicted, delta_days)
          if space_left_predicted <= threshold*1024*1024*1024:
            self.logger.error("The free disk space will be too low. " \
                              "(disk size: %.2f G, threshold: %s G)",
                                disk_size/(1024*1024*1024), threshold)
    else:
      self.logger.error("Free disk space low: remaining %.2f G (disk size: %.0f G, threshold: %.0f G).",
        free_space/(1024*1024*1024), disk_size/(1024*1024*1024), threshold)

  def test(self):
    return self._test(result_count=1, failure_amount=1)

  def anomaly(self):
    """
      Anomaly returns a TestResult instead of AnomalyResult because we don't
      want to call bang when there is a problem. Usually the problem won't be
      in the deployment of this instance but rather in one of the partition
      taking too much space.  This will need a human intervention.
    """
    return self._test(result_count=3, failure_amount=3)



import errno
import json
import logging
import os

from dateutil import parser
from .util import iter_logrotate_file_handle
from .util import iter_reverse_lines
from .util import JSONPromise

from zope.interface import implementer
from slapos.grid.promise import interface

@implementer(interface.IPromise)
class RunPromise(JSONPromise):

  def __init__(self, config):

    super(RunPromise, self).__init__(config)
    self.setPeriodicity(float(self.getConfig('frequency', 1)))
    self.netconf_log = self.getConfig('netconf-log')
    self.testing = self.getConfig('testing') == "True"

  def sense(self):

    if self.testing:
        self.logger.info("skipping promise")
        return

    for f in iter_logrotate_file_handle(self.netconf_log, 'rb'):
      for line in iter_reverse_lines(f):
        l = json.loads(line)
        alarm_notif = l.get('data', {}).get('notification', {}).get('alarm-notif', None)
        if alarm_notif and alarm_notif['fault-id'] == '9':
          if alarm_notif['is-cleared'] == 'false':
            affected_objects = alarm_notif.get('affected-objects', {})
            self.logger.error('VSWR alarm is on, affected objects are: %s', affected_objects)
            self.json_logger.info("Affected objects", extra={'data': affected_objects})
          else:
            self.logger.info('VSWR alarm is off')
          return
    self.logger.info('No VSWR alarm received')

  def test(self):
    """
      Called after sense() if the instance is still converging.
      Returns success or failure based on sense results.

      In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)


  def anomaly(self):
    """
      Called after sense() if the instance has finished converging.
      Returns success or failure based on sense results.
      Failure signals the instance has diverged.

      In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=1, failure_amount=1)
